# Copyright 2017 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Aruba generator."""

from typing import Union

from aerleon.lib import aclgenerator, policy
from aerleon.lib.nacaddr import IPv4, IPv6

_COMMENT_MARKER = '#'
_TERMINATOR_MARKER = '!'


class Error(aclgenerator.Error):
    """Base error class."""


class Term(aclgenerator.Term):
    """A single Aruba ACL term, mostly used for the __str__() method.

    Args:
      term: policy.Term object.
      filter_type: IP address version number.
    """

    _PLATFORM = 'aruba'

    _ANY_STR = 'any'
    _ALIAS_STR = 'alias'
    _IPV6_START_STR = 'ipv6'
    _NET_DEST_STR = 'netdestination'
    _NEGATOR = 'no'
    _SRC_NETDEST_SUF = '_src'
    _DST_NETDEST_SUF = '_dst'
    _NETWORK_STRING = 'network'
    _HOST_STRING = 'host'
    _USER_STR = 'user'
    _SOURCE_IS_USER_OPT_STR = 'source-is-user'
    _DESTINATION_IS_USER_OPT_STR = 'destination-is-user'
    _NEGATE_OPT_STR = 'negate'
    _IDENT = '  '

    _COMMENT_LINE_LENGTH = 70

    _ACTIONS = {
        'accept': 'permit',
        'deny': 'deny',
    }

    _PROTOCOL_MAP = {
        'icmp': 1,
        'gre': 47,
        'esp': 50,
    }

    def __init__(self, term: policy.Term, filter_type: str, verbose: bool = True) -> None:
        super().__init__(term)
        self.term = term
        self.filter_type = filter_type
        self.netdestinations = []
        self.verbose = verbose

    def __str__(self) -> str:
        netdestinations = []
        ret_str = []
        term_af = self.AF_MAP.get(self.filter_type)

        if self.term.verbatim:
            for next_verbatim in self.term.verbatim:
                if next_verbatim[0] == self._PLATFORM and next_verbatim[1]:
                    ret_str.append(f'{self._IDENT}{next_verbatim[1]}')

            return '\n'.join(t for t in ret_str if t)
        if self.verbose:
            comments = self.term.comment[:]

            if self.term.owner:
                comments.append(f'Owner: {self.term.owner}')

            if comments:
                for line in aclgenerator.WrapWords(comments, self._COMMENT_LINE_LENGTH):
                    ret_str.append(f'{self._IDENT}{_COMMENT_MARKER} {line}')

        src_addr_token = ''
        dst_addr_token = ''

        if self._SOURCE_IS_USER_OPT_STR in self.term.option:
            src_addr_token = self._USER_STR
        else:
            if self.term.source_address:
                src_addr = self.term.GetAddressOfVersion('source_address', term_af)
                if not src_addr:
                    return ''

                src_netdest_id = f'{self.term.name.lower()}{self._SRC_NETDEST_SUF}'
                src_addr_token = f'{self._ALIAS_STR} {src_netdest_id}'
                netdestinations.append(self._GenerateNetdest(src_netdest_id, src_addr, term_af))

            else:
                src_addr_token = self._ANY_STR

        if self._DESTINATION_IS_USER_OPT_STR in self.term.option:
            dst_addr_token = self._USER_STR
        else:
            if self.term.destination_address:
                dst_addr = self.term.GetAddressOfVersion('destination_address', term_af)
                if not dst_addr:
                    return ''

                dst_netdest_id = f'{self.term.name.lower()}{self._DST_NETDEST_SUF}'
                dst_addr_token = f'{self._ALIAS_STR} {dst_netdest_id}'
                netdestinations.append(self._GenerateNetdest(dst_netdest_id, dst_addr, term_af))
            else:
                dst_addr_token = self._ANY_STR

        dst_protocol_list = []
        if self.term.protocol:
            dst_protocol_list = self._GeneratePortTokens(
                self.term.protocol, self.term.destination_port
            )
        else:
            dst_protocol_list = [self._ANY_STR]

        for dst_port in dst_protocol_list:
            str_tok = [' ']

            if self._NEGATE_OPT_STR in self.term.option:
                str_tok.append(self._NEGATOR)

            if term_af == 6:
                str_tok.append(self._IPV6_START_STR)

            str_tok.append(src_addr_token)
            str_tok.append(dst_addr_token)

            str_tok.append(dst_port)
            str_tok.append(self._ACTIONS.get(self.term.action[0]))
            ret_str.append(' '.join(t for t in str_tok if t))

        self.netdestinations = netdestinations

        return '\n'.join(t for t in ret_str if t)

    def _GenerateNetdest(self, addr_netdestid: str, addresses: Union[IPv4, IPv6], af: int) -> str:
        """Generates the netdestinations text block.

        Args:
          addr_netdestid: netdestinations identifier.
          addresses: IP addresses.
          af: address family.
        Returns:
          A text block suitable for netdestinations in Aruba ACLs.
        """
        ret_str = []

        # Aruba does not use IP version identifier for IPv4.
        addr_family = '6' if af == 6 else ''

        ret_str.append(f'{self._NET_DEST_STR + addr_family} {addr_netdestid}')

        for address in addresses:
            ret_str.append(f'{self._IDENT}{self._GenerateNetworkOrHostTokens(address)}')

        ret_str.append(f'{_TERMINATOR_MARKER}\n')

        return '\n'.join(t for t in ret_str if t)

    def _GenerateNetworkOrHostTokens(self, address: Union[IPv4, IPv6]) -> str:
        """Generates the text block host or network identifier for netdestinations.

        Args:
          address: IP address.
        Returns:
          A string line using either 'host' or 'network', properly formatted for
          Aruba ACLs.
        """
        if address.num_addresses == 1:
            return f'{self._HOST_STRING} {address.network_address}'

        if address.version == 6:
            return f'{self._NETWORK_STRING} {address.network_address}/{address.prefixlen}'

        return f'{self._NETWORK_STRING} {address.network_address} {address.netmask}'

    def _GeneratePortTokens(self, protocols: list[str], ports: list[tuple[int, int]]) -> list[str]:
        """Generates string tokens for ports.

        Args:
          protocols: protocol to use (e.g. tcp, udp, etc.)
          ports: A list of tuples representing port ranges.
        Returns:
          A list of strings to be used as the port selector in Aruba ACLs.
        """
        ret_ports = []

        for protocol in protocols:
            if protocol in self._PROTOCOL_MAP:
                return [str(self._PROTOCOL_MAP[protocol])]
            for start_port, end_port in sorted(ports):
                ret_ports.append(
                    f'{protocol.lower()} {start_port}{f" {end_port!s}" if start_port != end_port else ""}'
                )
        return ret_ports


class Aruba(aclgenerator.ACLGenerator):
    """An Aruba policy object.

    This class takes a policy object and renders the output (via __str__ method)
    into a syntax which is understood by Aruba devices.

    Args:
      pol: policy.Policy object.
    """

    SUFFIX = '.aacl'
    _PLATFORM = 'aruba'

    _ACL_LINE_HEADER = 'ip access-list session'

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          Tuple containing both supported tokens and sub tokens.
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens -= {
            'destination_address_exclude',
            'icmp_type',
            'source_port',
            'source_address_exclude',
        }

        supported_sub_tokens.update(
            {
                'action': {
                    'accept',
                    'deny',
                },
                'option': {
                    'source-is-user',
                    'destination-is-user',
                    'negate',
                },
            }
        )

        del supported_sub_tokens['icmp_type']

        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        self.aruba_policies = []

        for header, terms in pol.filters:
            filter_name = header.FilterName(self._PLATFORM)
            filter_options = header.FilterOptions(self._PLATFORM)
            verbose = True
            if 'noverbose' in filter_options:
                filter_options.remove('noverbose')
                verbose = False

            filter_type = 'inet'
            if 'inet6' in filter_options:
                filter_type += '6'

            new_terms = []
            for term in terms:
                new_terms.append(Term(term, filter_type, verbose))

            self.aruba_policies.append((filter_name, new_terms, filter_type))

    def __str__(self) -> str:
        target = []

        target.extend(aclgenerator.AddRepositoryTags(f'{_COMMENT_MARKER} '))

        for filter_name, terms, _ in self.aruba_policies:
            netdestinations = []
            term_strings = []

            for term in terms:
                term_strings.append(str(term))
                netdestinations.extend(term.netdestinations)

            target.extend(netdestinations)
            target.append(f'{self._ACL_LINE_HEADER} {filter_name}')
            target.extend(term_strings)
            target.extend(_TERMINATOR_MARKER)

        if target:
            target.append('')

        return '\n'.join(target)



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Cisco IOS-XR filter renderer."""
from __future__ import annotations

from aerleon.lib import cisco
from aerleon.lib.policy import Term


class CiscoXR(cisco.Cisco):
    """A cisco policy object."""

    _PLATFORM = 'ciscoxr'
    _DEFAULT_PROTOCOL = 'ip'
    SUFFIX = '.xacl'
    _PROTO_INT = False

    def _AppendTargetByFilterType(self, filter_name: str, filter_type: str) -> list[str]:
        """Takes in the filter name and type and appends headers.

        Args:
          filter_name: Name of the current filter
          filter_type: Type of current filter

        Returns:
          list of strings
        """
        target = []
        if filter_type == 'inet6':
            target.append(f'no ipv6 access-list {filter_name}')
            target.append(f'ipv6 access-list {filter_name}')
        else:
            target.append(f'no ipv4 access-list {filter_name}')
            target.append(f'ipv4 access-list {filter_name}')
        return target

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {'next_ip'}

        return supported_tokens, supported_sub_tokens

    def _GetObjectGroupTerm(self, term: Term, verbose: bool = True) -> CiscoXRObjectGroupTerm:
        """Returns an ObjectGroupTerm object."""
        return CiscoXRObjectGroupTerm(term, platform=self._PLATFORM, verbose=verbose)


class CiscoXRObjectGroupTerm(cisco.ObjectGroupTerm):
    ALLOWED_PROTO_STRINGS = cisco.ExtendedTerm.ALLOWED_PROTO_STRINGS + ['pcp', 'esp']



"""Intermediate data models and classes for the front-end parsing phase.

Classes:

## RawPolicy, RawFilter, RawTerm, RawKV

These classes act as an intermediate representation between a raw file representation
and the main Policy data model.

## TValue, TComposite

These are recognizers (they have a method .Recognize(value)). They will parse value expressions
and extract normalized data from within.

## BuiltinRecognizer

This class uses recognizers to parse, validate and normalize all built-in fields in the Policy.
"""

from __future__ import annotations

import datetime
import enum
import re
import typing
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from aerleon.lib.policy_builder import (
        RawFilter,
        RawFilterHeader,
        RawPolicy,
        RawTarget,
        RawTerm,
    )


@dataclass
class RecognizerContext:
    policy: RawPolicy
    filter: RawFilter | None = None
    header: RawFilterHeader | None = None
    target: RawTarget | None = None
    term: RawTerm | None = None
    keyword: str | None = None
    value: str | None = None


@dataclass
class RecognizerKeywordResult:
    recognized: bool
    securityCritical: bool = False


@dataclass
class RecognizerValueResult:
    recognized: bool
    securityCritical: bool = False
    valueKV: dict | None = None


@dataclass
class RecognizerOptionResult:
    securityCritical: bool
    valueKV: dict | None = None


class TValue(enum.Enum):
    """Common recognizer types.

    These names are based loosely on the token names in Capirca's policy file
    parser. "WordString" is based on "STRING" and "AnyString" is based on "ANY".

    Generators that recognize extra keywords are encouraged
    but not required to use these types when implementing
    recognizeKeywordValue(). For the sake of consistency Generator
    authors should try to map their input to one of these value
    types, or a composition of value types (see TComposite).
    """

    AnyString = enum.auto()  # any value expression
    WordString = enum.auto()  # \w+
    Integer = enum.auto()  # \d+ string or YAML integer
    Hex = enum.auto()  # hex
    IntegerRange = enum.auto()  # Integer '-' Integer
    YearMonthDay = enum.auto()  # YYYY '-' MM '-' DD
    DSCP = enum.auto()  # DSCP traffic class
    DSCPRange = enum.auto()  # Range of DSCP traffic classes
    LogLimit = enum.auto()  # Integer '/' Str
    TargetResourceTuple = enum.auto()  # '(' Str ',' Str ')'

    def Recognize(self, value: typing.Any) -> str:
        """Match and parse the input value.

        Arguments:
            value: The input value.

        Returns: The input value after some parsing and normalization. For example,
            TValue.YearMonthDay would accept both a date object and a string of
            the form YYYY-MM-DD and return a date object in both cases.

        Raises:
            TypeError: If the input value does not match.
        """
        # .pol files allow strings with value "true" "True" "false" "False" and have no concept of a boolean type.
        if isinstance(value, bool):
            value = str(value).lower()

        if value is None:
            value = ""

        if self == TValue.AnyString:
            if not isinstance(value, str):
                raise TypeError("Expected string.")
            return value

        elif self == TValue.WordString:
            if not isinstance(value, str):
                raise TypeError("Expected string.")
            match = re.fullmatch(r'\w+([-_+.@/]\w*)*', value.strip())
            if match is None:
                raise TypeError("Expected value class 'String'.")
            return match[0]

        elif self == TValue.Integer:
            if isinstance(value, int):
                return value
            if not isinstance(value, str):
                raise TypeError("Expected integer or string.")
            match = re.fullmatch(r'\d+', value.strip())
            if match is None:
                raise TypeError("Expected value class 'Integer'.")
            return int(match[0])

        elif self == TValue.Hex:
            if not isinstance(value, str):
                raise TypeError("Expected string.")
            match = re.fullmatch(r'0x[a-fA-F0-9]+', value.strip())
            if match is None:
                raise TypeError("Expected value class 'Hex'.")
            return match[0]

        elif self == TValue.IntegerRange:
            if not isinstance(value, str):
                raise TypeError("Expected string.")
            match = re.fullmatch(r'(?P<start>\d+)\s*-\s*(?P<end>\d+)', value.strip())
            if match is None:
                raise TypeError("Expected value class 'IntegerRange'.")
            return f"{match['start']}-{match['end']}"

        elif self == TValue.YearMonthDay:
            if isinstance(value, datetime.date):
                return value
            if not isinstance(value, str):
                raise TypeError("Expected string or date.")
            match = re.fullmatch(
                r'(?P<year>\d+)\s*-\s*(?P<month>\d+)\s*-\s*(?P<day>\d+)', value.strip()
            )
            if match is None:
                raise TypeError("Expected value class 'YearMonthDay'.")
            return datetime.date(int(match['year']), int(match['month']), int(match['day']))

        elif self == TValue.DSCP:
            if isinstance(value, int):
                return str(value)
            if not isinstance(value, str):
                raise TypeError("Expected string or integer.")
            match = re.fullmatch(r'\d+', value.strip())
            if match:
                return match[0]
            match = re.fullmatch(
                r'\b((b[0-1]{6})|(af[1-4]{1}[1-3]{1})|(be)|(ef)|(cs[0-7]{1}))(?![\w-])\b',
                value.strip(),
            )
            if match:
                return match[0]
            raise TypeError("Expected value class 'DSCP'.")

        elif self == TValue.DSCPRange:
            try:
                return TValue.DSCP.Recognize(value)
            except TypeError:
                return '-'.join(
                    [TValue.DSCP.Recognize(subvalue) for subvalue in value.strip().split('-')]
                )

        elif self == TValue.LogLimit:
            if not isinstance(value, str):
                raise TypeError("Expected string.")
            match = re.fullmatch(r'(?P<frequency>\d+)\s*/\s*(?P<period>\w+)', value.strip())
            if match is None:
                raise TypeError("Expected value class 'LogLimit'.")
            return (match['frequency'], match['period'])

        elif self == TValue.TargetResourceTuple:
            if isinstance(value, list):
                if len(value) != 2:
                    raise TypeError("Expected list of length=2 or string.")
                first, second = value
            else:
                if not isinstance(value, str):
                    raise TypeError("Expected list or string.")
                match = re.fullmatch(r'\(\s*(.*)\s*,\s*(.*)\s*\)|(.*)\s*,\s*(.*)', value.strip())
                if match is None:
                    raise TypeError("Expected list or tuple expression.")
                first, second = (
                    (match[1], match[2]) if match[1] is not None else (match[3], match[4])
                )
            return (TValue.WordString.Recognize(first), TValue.WordString.Recognize(second))


class TComposition:
    of: typing.Any

    def Recognize(self, value: typing.Any) -> typing.Any:
        """Match and parse the input value."""
        pass


@dataclass
class TUnion(TComposition):
    """Recognizer that multiplexes a list of sub-recognizers.

    Attributes:
        of: A list of allowed types. The input must match one of these types.
    """

    of: list[TValue | TComposition]

    def Recognize(self, value):
        """Match and parse the input value using the list of sub-recognizers given in the 'of'
        class attribute.

        Arguments:
            value: The input value.

        Returns: The value processed by the first recognizer that matches it.

        Raises:
            TypeError: If none of the sub-recognizers match the input.
        """
        # Try each tokenizer on input
        for tokenizer in self.of:
            try:
                return tokenizer.Recognize(value)
            except TypeError:
                continue
        raise TypeError('Not recognized by this union type.')


@dataclass
class TList(TComposition):
    """Recognizer for list inputs.

    Attributes:
        of: The allowed type for members of this list.
        collapsible: Whether a list with a single item can be given directly.
    """

    of: TValue | TComposition
    collapsible: bool = False

    def Recognize(self, value: str) -> list[str]:
        """Match and parse the input value using the recognizer given in the 'of' class attribute.

        Arguments:
            value: The input value. Can be a list of values, a string (if space-separated list
            criteria are met), or a value that directly matches the 'of' recognizer.

        Returns: A list with each value processed by the recognizer given in attribute 'of'.

            If the 'collapsible' attribute is true, the input can be a value that matches the 'of'
            recognizer directly. For example, if 'of' is TValue.Integer, value=list(100) and
            value=100 are both acceptable inputs.

            If the 'of' attribute is a space-free TValue or a union of space-free TValues, a string
            input may be passed to this function. That string will be split by whitespace and the
            resulting list will be used as the input.

        Raises:
            TypeError: If any value in the list is not recognized by the 'of' attribute or
                if a non-list was provided when the collapsible list or space-separated list
                criteria are not met.
        """

        # TList accepts three modes of input:
        # (1) A list where each value is recognized by self.of
        # (2) A single value, dist or list recognized by self.of if collapsible=True.
        # (3) A string containing a space separated list of values (not supported for all value types).

        def _isSpaceFreeValue(value_type: TValue):
            return isinstance(value_type, TValue) and value_type not in (
                TValue.AnyString,
                TValue.TargetResourceTuple,
            )

        if value is None:
            value = []

        if isinstance(value, str) and _isSpaceFreeValue(self.of):
            return list(map(self.of.Recognize, value.split()))

        if (
            isinstance(value, str)
            and isinstance(self.of, TUnion)
            and all(_isSpaceFreeValue(value_type) for value_type in self.of.of)
        ):
            return list(map(self.of.Recognize, value.split()))

        if self.collapsible:
            try:
                return [self.of.Recognize(value)]
            except TypeError:
                pass

        if isinstance(value, list):
            return list(map(self.of.Recognize, value))
        else:
            if self.collapsible:
                raise TypeError("Expected a list or collapsed list.")
            else:
                raise TypeError("Expected a list.")


@dataclass
class TSection(TComposition):
    """Recognizer for dict inputs.

    Attributes:
        of: A list of rules for this section. Each rule is a 2-tuple with the first
            part containing the 'key rule' and the second part the 'value rule'.

            A 'key rule' can be:
                * A string literal.
                * A TValue recognizer.
                * A TUnion recognizer.

            A 'value rule' can be:
                * A TValue reconizer.
                * Any TComposition recognizer.

            A string literal key rule matches a key exactly. A TValue or TUnion key rule matches
            any key recognized by the given recognizer. A TValue or TComposition value rule matches
            any value recognized by the given recognizer.
    """

    of: list[tuple[str | TValue | TUnion, TValue | TComposition]]

    def Recognize(self, value: dict) -> dict[str, list[str]]:
        """Match and parse the input value using the list of rules given in the 'of' class attribute.
        See class docstring for more details on how to specify the rules list.

        Arguments:
            value: The input dict.

        Returns: A dict with the same keys as the input and with each value processed by the first
            matching value rule.

            Rules are processed in the order the appear in the 'of' list. In the event that a key
            rule matches but the value rule does not match, processing will continue through the
            list of rules.

        Raises:
            TypeError: If any key/value pair present in the input did not match any of the rules.
        """
        if value is None:
            value = {}
        if not isinstance(value, dict):
            raise TypeError("Expected a dictionary.")

        rules = self.of
        result = {}
        for keyword, keyword_value in value.items():
            last_error = None
            for key_rule, value_rule in rules:
                if isinstance(key_rule, str):
                    if keyword != key_rule:
                        continue
                else:
                    try:
                        keyword = key_rule.Recognize(keyword)
                    except TypeError:
                        continue
                try:
                    keyword_value = value_rule.Recognize(keyword_value)
                    result[keyword] = keyword_value
                    break
                except TypeError as e:
                    # Collect any TypeError but continue to search for another matching key_rule.
                    # If a subsequent key_rule / value_rule pair matches we can ignore the stored error.
                    # If no subsequent match is found we will raise the last value rule TypeError.
                    last_error = e
                    continue
            else:
                if last_error:
                    # A value rule TypeError was collected and no subsequent key rules matched.
                    raise last_error
                # None of the key rules matched `keyword`
                raise TypeError(f"Unexpected key {keyword} in section.")
        return result


# Pre-instantiate frequently used composite structures
TListStr = TList(of=TValue.WordString)
TListStrCollapsible = TList(of=TValue.WordString, collapsible=True)



# Copyright 2025 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NVIDIA NVUE API generator.

Generates ACL configurations for NVIDIA Cumulus Linux switches using the
NVUE (NVIDIA User Experience) REST API JSON format.

More information about NVUE:
https://docs.nvidia.com/networking-ethernet-software/cumulus-linux-514/System-Configuration/NVIDIA-User-Experience-NVUE/
"""

import itertools
import json
from typing import Optional

try:
    from absl import logging
except ImportError:
    import logging

from aerleon.lib import aclgenerator, policy

_ACTION_TABLE = {
    'accept': 'permit',
    'deny': 'deny',
    'reject': 'deny',
}

_PROTO_TABLE = {
    'icmp': 'icmp',
    'icmpv6': 'icmpv6',
    'tcp': 'tcp',
    'udp': 'udp',
    'esp': 'esp',
    'ah': 'ah',
    'gre': 'gre',
    'ipv6-icmp': 'icmpv6',
    'sctp': 'sctp',
}

_ICMP_TYPE_TABLE = {
    'echo-reply': 'echo-reply',
    'echo-request': 'echo-request',
    'unreachable': 'destination-unreachable',
    'time-exceeded': 'time-exceeded',
}


class Error(aclgenerator.Error):
    """Base error class for NVUE generator."""


class NvueApiError(Error):
    """Raised when there's an error generating NVUE API config."""


class UnsupportedNvueFilterError(Error):
    """Raised when we can't create the requested acl."""


class Term(aclgenerator.Term):
    """Representation of an individual NVUE ACL rule."""

    _PLATFORM = 'nvueapi'

    def __init__(self, term: policy.Term, address_family: str = 'ipv4'):
        """Initialize NVUE term.

        Args:
            term: The policy term object
            address_family: 'ipv4' or 'ipv6'
        """
        super().__init__(term)
        self.term = term
        self.address_family = address_family

    def __str__(self) -> str:
        """Convert term to NVUE JSON rule format."""
        # This should not be called directly - use GenerateRules() instead
        # which handles the cartesian product of multiple addresses
        raise NotImplementedError("Use GenerateRules() method instead")

    def GenerateRules(self) -> list[dict]:
        """Generate NVUE rules handling multiple addresses.

        Since NVUE only supports single CIDR prefixes per rule (like iptables),
        we need to create separate rules for each combination of addresses.

        Returns:
            List of rule dictionaries
        """
        rules = []

        # Get address lists for this address family
        source_addrs = self._GetFilteredAddresses(self.term.source_address)
        dest_addrs = self._GetFilteredAddresses(self.term.destination_address)

        # If no addresses specified, use empty list to generate one rule
        if not source_addrs:
            source_addrs = [None]
        if not dest_addrs:
            dest_addrs = [None]

        # Get other rule components
        protocols = self._GetProtocols()
        dest_ports = self._GetPorts(self.term.destination_port)
        source_ports = self._GetPorts(self.term.source_port)
        icmp_types = self._GetIcmpTypes()

        # Generate cartesian product of all combinations using itertools
        for src_addr, dst_addr, protocol, dest_port, source_port, icmp_type in itertools.product(
            source_addrs, dest_addrs, protocols, dest_ports, source_ports, icmp_types
        ):
            rule_dict = self._CreateSingleRule(
                src_addr, dst_addr, protocol, dest_port, source_port, icmp_type
            )
            if rule_dict:  # Only add non-empty rules
                rules.append(rule_dict)

        return rules

    def _GetFilteredAddresses(self, addresses) -> list[str]:
        """Get addresses filtered by address family.

        Args:
            addresses: List of address objects to filter

        Returns:
            List of address strings matching the current address family
        """
        if not addresses:
            return []

        filtered = []
        for addr in addresses:
            if addr.version == 4 and self.address_family == 'ipv4':
                filtered.append(str(addr))
            elif addr.version == 6 and self.address_family == 'ipv6':
                filtered.append(str(addr))
        return filtered

    def _GetProtocols(self) -> list[Optional[str]]:
        """Get protocol list, translating Aerleon protocol names to NVUE format.

        Returns:
            List of protocol names in NVUE format, or [None] if no protocols specified
        """
        if not self.term.protocol:
            return [None]
        return [_PROTO_TABLE.get(p, p) for p in self.term.protocol]

    def _GetPorts(self, port_list) -> list[Optional[str]]:
        """Get port list in NVUE format (porta:portz).

        Args:
            port_list: List of port ranges from the policy term

        Returns:
            List of port strings in NVUE format, or [None] if no ports specified
        """
        if not port_list:
            return [None]

        ports = []
        for port_range in port_list:
            if port_range[0] == port_range[1]:
                ports.append(str(port_range[0]))
            else:
                # NVUE uses porta:portz format for ranges
                ports.append(f"{port_range[0]}:{port_range[1]}")
        return ports

    def _GetIcmpTypes(self) -> list[Optional[str]]:
        """Get ICMP type list, converting from Aerleon format to iptables format.

        Returns:
            List of ICMP type strings in iptables format, or [None] if no ICMP types specified
        """
        if not self.term.icmp_type:
            return [None]

        return [_ICMP_TYPE_TABLE.get(icmp_type, icmp_type) for icmp_type in self.term.icmp_type]

    def _CreateSingleRule(
        self,
        src_addr: Optional[str],
        dst_addr: Optional[str],
        protocol: Optional[str],
        dest_port: Optional[str],
        source_port: Optional[str],
        icmp_type: Optional[str],
    ) -> dict:
        """Create a single NVUE rule from the given parameters.

        Args:
            src_addr: Source IP address/network
            dst_addr: Destination IP address/network
            protocol: Protocol name
            dest_port: Destination port or port range
            source_port: Source port or port range
            icmp_type: ICMP type name

        Returns:
            Dictionary representing a single NVUE rule
        """
        rule_dict = {}

        # Set action - NVUE uses object format with empty dict values
        action_dict = {}

        # Handle primary action
        if self.term.action:
            action = self.term.action[0]
            nvue_action = _ACTION_TABLE.get(action, action)
            action_dict[nvue_action] = {}

        # Add logging if specified
        if self.term.logging:
            action_dict['log'] = {}

        # If no actions specified, default to permit
        if not action_dict:
            action_dict['permit'] = {}

        rule_dict['action'] = action_dict

        # Set remark/comment
        if self.term.comment:
            rule_dict['remark'] = ' '.join(self.term.comment)

        # Set match conditions
        match_dict = {}
        ip_dict = {}

        # Protocol matching
        if protocol:
            ip_dict['protocol'] = protocol

        # Source IP matching
        if src_addr:
            ip_dict['source-ip'] = src_addr

        # Destination IP matching
        if dst_addr:
            ip_dict['dest-ip'] = dst_addr

        # Destination port matching
        if dest_port:
            ip_dict['dest-port'] = dest_port

        # Source port matching
        if source_port:
            ip_dict['source-port'] = source_port

        # ICMP type matching
        if icmp_type and protocol == 'icmp' and self.address_family == 'ipv4':
            ip_dict['icmp-type'] = icmp_type
        elif icmp_type and protocol == 'icmpv6' and self.address_family == 'ipv6':
            ip_dict['icmpv6-type'] = icmp_type

        # TCP state matching for NVUE
        tcp_dict = {}
        if protocol == 'tcp' and self.term.option:
            for option in self.term.option:
                if option in ['established', 'tcp-established']:
                    tcp_dict['state'] = 'established'
                    break

        # Add IP match conditions if any exist
        if ip_dict:
            match_dict['ip'] = ip_dict

        # Add TCP conditions if any exist (at same level as ip)
        if tcp_dict:
            match_dict['tcp'] = tcp_dict

        # Add match conditions if any exist
        if match_dict:
            rule_dict['match'] = match_dict

        return rule_dict


class NvueApi(aclgenerator.ACLGenerator):
    """NVUE API generator class."""

    _PLATFORM = 'nvueapi'
    _DEFAULT_PROTOCOL = 'ip'
    SUFFIX = '.nvueapi.json'
    SUPPORTED_AF = {'inet', 'inet6'}
    SUPPORTED_TARGETS = frozenset(['nvueapi'])
    WARN_IF_UNSUPPORTED = frozenset(['translated', 'stateless_reply', 'counter', 'policer'])

    def __init__(self, policy_obj: policy.Policy, exp_info: int) -> None:
        """Initialize NVUE API generator.

        Args:
            policy_obj: The policy object
            exp_info: Expiration info integer
        """
        self.nvue_policies = []
        self.address_family = None
        super().__init__(policy_obj, exp_info)

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        """Translate policy to NVUE format.

        Args:
            pol: The policy object
            exp_info: Expiration info
        """
        self.nvue_policies = []

        for header, terms in pol.filters:
            if self._PLATFORM not in [x.platform for x in header.target]:
                continue

            filter_options = header.FilterOptions(self._PLATFORM)[1]
            filter_name = header.FilterName(self._PLATFORM)

            # Determine address family from filter options
            # NVUE supports ipv4, ipv6, or mac, but Aerleon only supports IP-based ACLs
            if 'ipv6' in filter_options:
                self.address_family = 'ipv6'
            elif 'inet6' in filter_options:
                self.address_family = 'ipv6'
            elif 'mac' in filter_options:
                raise UnsupportedNvueFilterError(
                    'NVUE MAC ACLs are not supported by Aerleon. '
                    'Aerleon only supports IP-based access control.'
                )
            elif 'mixed' in filter_options:
                raise UnsupportedNvueFilterError(
                    'NVUE does not support mixed address family ACLs. '
                    'Use separate ipv4 and ipv6 ACLs instead.'
                )
            else:
                self.address_family = 'ipv4'

            # Build NVUE ACL structure
            acl_rules = {}
            rule_number = 10  # Start rule numbering at 10, increment by 10

            for term in terms:
                if term.expiration:
                    if term.expiration <= exp_info:
                        logging.info(
                            'INFO: Term %s in policy %s expires ' 'in less than two weeks.',
                            term.name,
                            filter_name,
                        )
                    if term.expiration <= exp_info:
                        logging.warning(
                            'WARNING: Term %s in policy %s is expired and '
                            'will not be rendered.',
                            term.name,
                            filter_name,
                        )
                        continue

                # Handle address family filtering
                if self.address_family == 'ipv4':
                    # Skip terms that only have IPv6 addresses
                    has_ipv4 = self._HasIPv4(term)
                    if not has_ipv4:
                        continue
                elif self.address_family == 'ipv6':
                    # Skip terms that only have IPv4 addresses
                    has_ipv6 = self._HasIPv6(term)
                    if not has_ipv6:
                        continue

                # Generate rules for this term (may be multiple due to address expansion)
                nvue_term = Term(term, self.address_family)
                term_rules = nvue_term.GenerateRules()

                for rule_dict in term_rules:
                    if rule_dict:  # Only add non-empty rules
                        acl_rules[str(rule_number)] = rule_dict
                        rule_number += 10

            # Create the full ACL structure
            acl_config = {'acl': {filter_name: {'type': self.address_family, 'rule': acl_rules}}}

            self.nvue_policies.append((filter_name, acl_config))

    def _HasIPv4(self, term: policy.Term) -> bool:
        """Check if term has IPv4 addresses.

        Args:
            term: Policy term to check

        Returns:
            True if term has IPv4 addresses or no addresses, False otherwise
        """
        for addr_list in [term.source_address, term.destination_address]:
            if addr_list:
                for addr in addr_list:
                    if addr.version == 4:
                        return True
        return True  # Allow terms without addresses

    def _HasIPv6(self, term: policy.Term) -> bool:
        """Check if term has IPv6 addresses.

        Args:
            term: Policy term to check

        Returns:
            True if term has IPv6 addresses or no addresses, False otherwise
        """
        for addr_list in [term.source_address, term.destination_address]:
            if addr_list:
                for addr in addr_list:
                    if addr.version == 6:
                        return True
        return True  # Allow terms without addresses

    def __str__(self) -> str:
        """Return the NVUE configuration as a JSON string."""
        if not self.nvue_policies:
            return ''

        # Combine all ACL configurations
        acl_config = {}

        for _, policy_config in self.nvue_policies:
            acl_config.update(policy_config['acl'])

        # Return ACL configuration directly without 'set:' wrapper
        nvue_config = {'acl': acl_config}

        return json.dumps(nvue_config, indent=2)

    def _BuildTokens(self):
        """Build supported tokens for the NVUE platform.

        Returns:
            Tuple of (supported_tokens, supported_sub_tokens) sets
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        # Add NVUE-specific tokens
        supported_tokens |= {
            'action',
            'comment',
            'destination_address',
            'destination_port',
            'icmp_type',
            'logging',
            'name',
            'option',  # For TCP established state
            'protocol',
            'source_address',
            'source_port',
        }

        # Remove unsupported tokens
        supported_tokens -= {
            'verbatim',  # NVUE doesn't support verbatim rules
            'icmp_code',  # Not implemented yet
            'destination_address_exclude',  # NVUE doesn't support address exclusion
            'source_address_exclude',  # NVUE doesn't support address exclusion
            'stateless_reply',  # Not implementing stateless reply functionality
            'translated',  # Not implementing NAT functionality
        }

        # NVUE-specific sub-tokens
        supported_sub_tokens = {
            'action': {'accept', 'deny', 'reject'},
        }

        return supported_tokens, supported_sub_tokens



# Copyright 2021 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Openconfig yang ACL generator.

More information about the Openconfig ACL model schema:
http://ops.openconfig.net/branches/models/master/openconfig-acl.html
"""

import copy
import json
from collections import defaultdict
from typing import Any, DefaultDict, TypedDict, Union

from absl import logging

from aerleon.lib import aclgenerator, policy


class Error(aclgenerator.Error):
    """Generic error class."""


class OcFirewallError(Error):
    """Raised with problems in formatting for OpenConfig firewall."""


class ExceededAttributeCountError(Error):
    """Raised when the total attribute count of a policy is above the maximum."""


class TcpEstablishedWithNonTcpError(Error):
    """Raised when the TCP established option is set with a non TCP protocol."""


# Graceful handling of dict heierarchy for OpenConfig JSON.
def RecursiveDict() -> DefaultDict[Any, Any]:
    return defaultdict(RecursiveDict)


TransportConfig = TypedDict(
    "TransportConfig",
    {
        "source-port": Union[int, str],
        "destination-port": Union[int, str],
        "detail-mode": str,
        "builtin-detail": str,
    },
)


class Transport(TypedDict):
    transport: TransportConfig


IPConfig = TypedDict(
    "IPConfig", {"source-address": str, "destination-address": str, "protocol": int}
)


class IP(TypedDict):
    config: IPConfig


ActionConfig = TypedDict("ActionConfig", {"forwarding-action": str})


class Action(TypedDict):
    config: ActionConfig


ACLEntry = TypedDict(
    "ACLEntry",
    {"sequence-id": int, "actions": Action, "ipv4": IP, "ipv6": IP, "transport": Transport},
)
aclEntries = TypedDict("aclEntries", {"acl-entry": list[ACLEntry]})


class ACLSetConfig(TypedDict):
    name: str
    type: str


ACLSet = TypedDict(
    "ACLSet", {"acl-entries": aclEntries, "config": ACLSetConfig, "name": str, "type": str}
)


class Term(aclgenerator.Term):
    """Creates the term for the OpenConfig firewall."""

    ACTION_MAP = {'accept': 'ACCEPT', 'deny': 'DROP', 'reject': 'REJECT'}

    # OpenConfig ip-protocols always will resolve to an 8-bit int, but these
    # common names are more convenient in a policy file.
    _ALLOW_PROTO_NAME = frozenset(['tcp', 'udp', 'icmp', 'esp', 'ah', 'ipip', 'sctp'])

    AF_RENAME = {
        4: 'ipv4',
        6: 'ipv6',
    }

    def __init__(self, term: policy.Term, inet_version: str = 'inet') -> None:
        super().__init__(term)
        self.term = term
        self.inet_version = inet_version

        # Combine (flatten) addresses with their exclusions into a resulting
        # flattened_saddr, flattened_daddr, flattened_addr.
        self.term.FlattenAll()

    def __str__(self) -> None:
        """Convert term to a string."""
        rules = self.ConvertToDict()
        json.dumps(rules, indent=2)

    def _tcp_established(self) -> dict[str, str]:
        """Return's openconfig TCP_ESTABLISHED configuration.

        Other vendors (eg. SONiC) have slighly different implementations,
        This function permits inheritance."""
        return {'detail-mode': 'BUILTIN', 'builtin-detail': "TCP_ESTABLISHED"}

    def ConvertToDict(self, filter_options: list[str]) -> list[ACLEntry]:
        """Convert term to a dictionary.

        This is used to get a dictionary describing this term which can be
        output easily as an Openconfig JSON blob. It represents an "acl-entry"
        message from the OpenConfig ACL schema.

        Returns:
          A list of dictionaries that contains all fields necessary to create or
          update a OpenConfig acl-entry.
        """
        self.term_dict = RecursiveDict()

        # Rules will hold all exploded acl-entry dictionaries.
        rules = []

        # Convert the integer to the proper openconfig schema name str, ipv4/ipv6.
        term_af = self.AF_MAP.get(self.inet_version)
        family = self.AF_RENAME[term_af]

        self.SetName(self.term.name)

        # Action
        self.SetAction(filter_options)

        # Ballot fatigue handling for 'any'.
        saddrs = self.term.GetAddressOfVersion('flattened_saddr', term_af)
        if not saddrs:
            saddrs = ['any']

        daddrs = self.term.GetAddressOfVersion('flattened_daddr', term_af)
        if not daddrs:
            daddrs = ['any']

        sports = self.term.source_port
        if not sports:
            sports = [(0, 0)]

        dports = self.term.destination_port
        if not dports:
            dports = [(0, 0)]

        protos = self.term.protocol
        if not protos:
            protos = ['none']

        self.term_dict = copy.deepcopy(self.term_dict)

        if self.term.comment:
            self.SetComments(self.term.comment)

        # Options
        self.SetOptions(family, filter_options)

        # Source Addresses
        for saddr in saddrs:
            if saddr != 'any':
                self.SetSourceAddress(family, str(saddr), filter_options)

            # Destination Addresses
            for daddr in daddrs:
                if daddr != 'any':
                    self.SetDestAddress(family, str(daddr), filter_options)

                # Source Port
                for start, end in sports:
                    # 'any' starts and ends with zero.
                    if not start == end == 0:
                        self.SetSourcePorts(start, end, filter_options)

                    # Destination Port
                    for start, end in dports:
                        if not start == end == 0:
                            self.SetDestPorts(start, end, filter_options)

                        # Protocol
                        for proto in protos:
                            if isinstance(proto, str):
                                if proto != 'none':
                                    try:
                                        proto_num = self.PROTO_MAP[proto]
                                    except KeyError:
                                        raise OcFirewallError(
                                            'Protocol %s unknown. Use an integer.', proto
                                        )
                                    self.SetProtocol(family, proto_num, filter_options)
                            else:
                                self.SetProtocol(family, proto, filter_options)

                            # This is the business end of ace explosion.
                            # A dict is a reference type, so deepcopy is actually required.
                            rules.append(copy.deepcopy(self.term_dict))

        return rules

    def SetName(self, name: str) -> None:
        pass

    def SetAction(self, filter_options: list[str]) -> None:
        action = self.ACTION_MAP[self.term.action[0]]
        self.term_dict['actions'] = {}
        self.term_dict['actions']['config'] = {}
        self.term_dict['actions']['config']['forwarding-action'] = action

    def SetComments(self, comments: list[str]) -> None:
        pass

    def SetOptions(self, family: str, filter_options: list[str]) -> None:
        # options, 'family' unused
        if self.term.option:
            if 'tcp-established' in self.term.option:
                if self.term.protocol != ['tcp']:
                    raise TcpEstablishedWithNonTcpError(
                        f'tcp-established can only be used with tcp protocol in term {self.term.name}'
                    )
                self.term_dict['transport']['config'].update(self._tcp_established())

    def SetSourceAddress(self, family: str, saddr: str, filter_options: list[str]) -> None:
        self.term_dict[family]['config']['source-address'] = saddr

    def SetDestAddress(self, family: str, daddr: str, filter_options: list[str]) -> None:
        self.term_dict[family]['config']['destination-address'] = daddr

    def SetSourcePorts(self, start: int, end: int, filter_options: list[str]) -> None:
        if start == end:
            self.term_dict['transport']['config']['source-port'] = start
        else:
            self.term_dict['transport']['config']['source-port'] = '%d..%d' % (
                start,
                end,
            )

    def SetDestPorts(self, start: int, end: int, filter_options: list[str]) -> None:
        if start == end:
            self.term_dict['transport']['config']['destination-port'] = start
        else:
            self.term_dict['transport']['config']['destination-port'] = '%d..%d' % (
                start,
                end,
            )

    def SetProtocol(self, family: str, protocol: int, filter_options: list[str]) -> None:
        self.term_dict[family]['config']['protocol'] = protocol


class OpenConfig(aclgenerator.ACLGenerator):
    """A OpenConfig firewall policy object."""

    _PLATFORM = 'openconfig'
    SUFFIX = '.oacl'
    _SUPPORTED_AF = frozenset(('inet', 'inet6', 'mixed'))
    FAMILY_MAP = {'mixed': 'ACL_MIXED', 'inet6': 'ACL_IPV6', 'inet': 'ACL_IPV4'}
    _TERM = Term

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        # Remove unsupported things
        supported_tokens -= {'platform', 'platform_exclude', 'icmp-type', 'verbatim'}

        # OpenConfig ACL model only supports these three forwarding actions.
        supported_sub_tokens['action'] = {'accept', 'deny', 'reject'}

        supported_sub_tokens.update(
            {
                'option': {
                    'tcp-established',
                }
            }
        )

        return supported_tokens, supported_sub_tokens

    def _InitACLSet(self) -> None:
        """Initialize self.acl_sets with proper Typing"""
        self.acl_sets: list[ACLSet] = []

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int) -> None:
        self.total_rule_count = 0
        self._InitACLSet()

        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)

            # Options are anything after the platform name in the target message of
            # the policy header, [1:].

            # Get the address family if set.
            address_family = 'inet'
            for i in self._SUPPORTED_AF:
                if i in filter_options:
                    address_family = i
                    filter_options.remove(i)
            self._TranslateTerms(
                terms, address_family, filter_name, header.comment, filter_options
            )

        logging.info('Total rule count of policy %s is: %d', filter_name, self.total_rule_count)

    def _TranslateTerms(
        self,
        terms: list[Term],
        address_family: str,
        filter_name: str,
        hdr_comments: list[str],
        filter_options: list[str],
    ) -> None:
        """
        Factor out the translation of terms, such that it can be overridden by subclasses
        """
        oc_acl_entries: list[ACLEntry] = []

        for term in terms:
            # Handle mixed for each indvidual term as inet and inet6.
            # inet/inet6 are treated the same.
            term_address_families = []
            if address_family == 'mixed':
                term_address_families = ['inet', 'inet6']
            else:
                term_address_families = [address_family]
            for term_af in term_address_families:
                t = self._TERM(term, term_af)
                for rule in t.ConvertToDict(filter_options):
                    self.total_rule_count += 1
                    rule['sequence-id'] = self.total_rule_count * 5
                    oc_acl_entries.append(rule)
        oc_type = self.FAMILY_MAP[address_family]
        oc_acl_set = {
            "acl-entries": {"acl-entry": oc_acl_entries},
            "config": {"name": filter_name, "type": oc_type},
            "name": filter_name,
            "type": oc_type,
        }
        self.acl_sets.append(oc_acl_set)

    def __str__(self) -> str:
        out = '%s\n\n' % (
            json.dumps(self.acl_sets, indent=2, separators=(',', ': '), sort_keys=True)
        )

        return out



"""Plugin Supervisor.

## Configuring the Plug-in Supervisor

PluginSupervisor.Start() will by default search for and load any installed
packages with a plugin entrypoint matching "aerleon.plugin".

## Plugin Lifecycle

Aerleon does "interrogate" each plug-in on start-up. During the interrogation
process, Aerleon determines:
* Whether this plug-in is compatible with the current version of Aerleon.
* What type of plug-in this is.
Aerleon also gives the plug-in an opportunity to detect an incompatible version
(the plug-in author may have more specific information about what Aerleon
versions it can support).

Incompatible plug-ins will be ignored with a warning.

All compatible plug-ins are then asked for internal resources which are
registered to internal tables. For Generator plug-ins, each plug-in is
asked for a table mapping a target name (e.g. "iptables") to a provided
class constructor. Aerleon will consult these mapping tables later during
run time.

For Generator classes, Aerleon will initialize these classes as needed.
The interface between Aerleon and Generator classes is not part of the
plug-in lifecycle or the PluginSupervisor's function.

## Authoring Plug-ins

Plug-in authors should consult the docstring for plugin.py. This file provides
documentation of the requirements for plug-ins as well as a suggested base
class for any Aerleon plug-in to use (BasePlugin).

## Debugging

Plug-in authors may observe that this module exposes its internal state
directly on the module for transparency. These names should be treated as
undocumented and subject to change.
"""

from __future__ import annotations

import importlib.util
import pathlib
import sys
from dataclasses import dataclass
from importlib import import_module

from aerleon.lib.plugin import SystemMetadata

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points, version
else:
    from importlib.metadata import entry_points, version

from absl import logging

from aerleon.lib import plugin
from aerleon.lib.aclgenerator import ACLGenerator

__all__ = [
    "PluginSupervisor",
    "PluginSupervisorConfiguration",
]


class _PluginSupervisor:
    is_setup: bool
    plugins: list[tuple]
    generators: dict[str, type[ACLGenerator]]

    def __init__(self) -> None:
        self.is_setup = False

    def Start(self, config: PluginSupervisorConfiguration | None = None) -> None:
        setup = _PluginSetup(config)
        self.plugins, self.generators = setup.plugins, setup.generators
        self.is_setup = True


__doc_PluginSupervisor__ = """Singleton PluginSupervisor instance."""
PluginSupervisor = _PluginSupervisor()

__doc_SYSTEM_METADATA__ = """Public module constant system metadata."""
SYSTEM_METADATA: SystemMetadata = SystemMetadata(engine_version=version("aerleon"))

__doc_BUILTIN_PLUGINS__ = (
    """Built-in plugins included with this project. These will always be loaded.""",
)
BUILTIN_GENERATORS: list[tuple] = [
    # fmt: off
    #Target                  Module                              Constructor
    ('juniper',              'aerleon.lib.juniper',              'Juniper'),
    ('juniperevo',           'aerleon.lib.juniperevo',           'JuniperEvo'),
    ('msmpc',                'aerleon.lib.junipermsmpc',         'JuniperMSMPC'),
    ('srx',                  'aerleon.lib.junipersrx',           'JuniperSRX'),
    ('cisco',                'aerleon.lib.cisco',                'Cisco'),
    ('ciscoasa',             'aerleon.lib.ciscoasa',             'CiscoASA'),
    ('aruba',                'aerleon.lib.aruba',                'Aruba'),
    ('brocade',              'aerleon.lib.brocade',              'Brocade'),
    ('arista',               'aerleon.lib.arista',               'Arista'),
    ('arista_tp',            'aerleon.lib.arista_tp',            'AristaTrafficPolicy'),
    ('ipset',                'aerleon.lib.ipset',                'Ipset'),
    ('iptables',             'aerleon.lib.iptables',             'Iptables'),
    ('nsxv',                 'aerleon.lib.nsxv',                 'Nsxv'),
    ('nsxt',                 'aerleon.lib.nsxt',                 'Nsxt'),
    ('openconfig',           'aerleon.lib.openconfig',           'OpenConfig'),
    ('sonic',                'aerleon.lib.sonic',                'SONiC'),
    ('speedway',             'aerleon.lib.speedway',             'Speedway'),
    ('pcap',                 'aerleon.lib.pcap',                 'PcapFilter'),
    ('pcap',                 'aerleon.lib.pcap',                 'PcapFilter'),
    ('packetfilter',         'aerleon.lib.packetfilter',         'PacketFilter'),
    ('windows_advfirewall',  'aerleon.lib.windows_advfirewall',  'WindowsAdvFirewall'),
    ('srxlo',                'aerleon.lib.srxlo',                'SRXlo'),
    ('cisconx',              'aerleon.lib.cisconx',              'CiscoNX'),
    ('ciscoxr',              'aerleon.lib.ciscoxr',              'CiscoXR'),
    ('nftables',             'aerleon.lib.nftables',             'Nftables'),
    ('nokiasrl',             'aerleon.lib.nokiasrl',             'NokiaSRLinux'),
    ('nvueapi',              'aerleon.lib.nvueapi',              'NvueApi'),
    ('gce',                  'aerleon.lib.gce',                  'GCE'),
    ('gcp_hf',               'aerleon.lib.gcp_hf',               'HierarchicalFirewall'),
    ('paloalto',             'aerleon.lib.paloaltofw',           'PaloAltoFW'),
    ('cloudarmor',           'aerleon.lib.cloudarmor',           'CloudArmor'),
    ('k8s',                  'aerleon.lib.k8s',                  'K8s'),
    ('fortigate',            'aerleon.lib.fortigate',            'Fortigate'),
    ('proxmox',              'aerleon.lib.proxmox',              'Proxmox'),
    # fmt: on
]


class PluginSetupCollisionError(Exception):
    """A plugin was already loaded for the same firewall platform."""


@dataclass
class PluginSupervisorConfiguration:
    """
    Attributes:
        disable_discovery: Plugin discovery via pip will not be performed if True. Consider
            this setting if plugin discovery is not needed and problematic modules are present
            in this pip installation or discovery is taking a long time.

        disable_plugin: A list of discovered plugins to ignore, given by plugin name. Consider
            this setting if one or more specific plugins are causing problems.

        disable_builtin: A list of built-in generators to ignore, given by module name (e.g.
            'juniper'). Use this setting if you plan to use an alternative generator for
            a firewall platform with built-in support.

        include_path: A list of plug-ins to include, given by package name. Through this
            option, plug-in packages can be included if they are present in the same
            Python installation but not registered through the normal Python plugin
            entrypoint system. Plug-ins included this way are expeted to implement a
            top-level function called "AerleonPlugin" that returns an instance of
            aerleon.lib.Plugin .
    """

    disable_discovery: bool = False
    disable_plugin: list[str] | None = None
    disable_builtin: list[str] | None = None
    include_path: list[list[str]] | None = None


class _PluginSetup:
    """
    Discover, load, interrogate and initialize all available plugins.

    Attributes:
        generators: All generators loaded from plugins.
        plugins: All loaded plugins.
        disable_discovery: See PluginSupervisorConfiguration.
        disable_plugin: See PluginSupervisorConfiguration.
        disable_builtin: See PluginSupervisorConfiguration.
        include_path: See PluginSupervisorConfiguration.
    """

    disable_discovery: bool = False
    disable_plugin: list[str] | None = None
    disable_builtin: list[str] | None = None
    include_path: list[list[str]] | None = None

    def __init__(self, config: PluginSupervisorConfiguration | None = None) -> None:
        """Initialize self.generators, self.plugins."""

        self.generators: dict[str, type[ACLGenerator]] = {}
        self.plugins = []

        # Apply configuration if provided
        if config is not None:
            self.disable_discovery = getattr(config, 'disable_discovery', False)
            self.disable_plugin = getattr(config, 'disable_plugin', None)
            self.disable_builtin = getattr(config, 'disable_builtin', None)
            self.include_path = getattr(config, 'include_path', None)

        # Initialize generator list with built-in generators
        self.generators.update(self._CollectBuiltinGenerators(BUILTIN_GENERATORS))

        # Collect plugins from various sources
        loaded_plugins = []
        if not self.disable_discovery:
            loaded_plugins.extend(self._CollectEntrypointPlugins())
        if self.include_path:
            loaded_plugins.extend(self._CollectPluginsByPath())

        # Attempt to load, initialize, interrogate, and register generators from each plugin
        for plugin_name, loaded_plugin in loaded_plugins:
            # Initialize entrypoint class or function and request metadata
            try:
                plugin_instance: plugin.BasePlugin = loaded_plugin()
                metadata = plugin_instance.RequestMetadata(SYSTEM_METADATA)
                if not isinstance(metadata, plugin.PluginMetadata) or any(
                    filter(
                        lambda c: not isinstance(c, plugin.PluginCapability), metadata.capabilities
                    )
                ):
                    logging.warning(
                        f"Ignoring plugin {plugin_name}: unrecognized plugin metadata format. engine_version={SYSTEM_METADATA.engine_version}"  # noqa E501
                    )
                    continue
                self.plugins.append((loaded_plugin, metadata))
            except plugin.PluginCompatibilityError as exception:
                logging.warning(
                    f"Ignoring plugin {plugin_name}: Aerleon version not supported by plugin. engine_version={SYSTEM_METADATA.engine_version}",  # noqa E501
                    exc_info=exception,
                )
                continue

            # Register generators
            if plugin.PluginCapability.GENERATOR not in metadata.capabilities:
                logging.warning(
                    f"Ignoring plugin {plugin_name}: GENERATOR capability not found. engine_version={SYSTEM_METADATA.engine_version}"  # noqa E501
                )
                continue

            try:
                plugin_generator_items = plugin_instance.generators
            except Exception as exception:
                logging.warning(
                    f"Plugin {plugin_name} crashed while registering generator.",
                    exc_info=exception,
                )
                continue

            for target, constructor in plugin_generator_items.items():
                found = self.generators.setdefault(target, constructor)
                if found != constructor:
                    # collision: two plugins claim to provide support for the same platform
                    raise PluginSetupCollisionError(
                        f"Plugin misconfiguration: more than one plugin is installed for {target}. Plugin 1: {found}. Plugin 2: {constructor}."  # noqa E501
                    )

        logging.info(f"{len(self.plugins)} plugins active.")
        logging.info(f"{len(self.generators)} generators registered.")

    def _CollectEntrypointPlugins(self) -> list[tuple[str, plugin.BasePlugin]]:
        """Locate and import modules using entrypoint discovery."""
        loaded_plugins: list[tuple[str, plugin.BasePlugin]] = []
        for ep_plugin in entry_points(group='aerleon.plugin'):
            if self.disable_plugin and ep_plugin.name in self.disable_plugin:
                continue
            try:
                loaded_plugins.append((ep_plugin.name, ep_plugin.load()))
            except Exception as exception:
                logging.warning(f"Failed to load plugin {ep_plugin.name}", exc_info=exception)
                continue
        return loaded_plugins

    def _CollectPluginsByPath(self):
        """Import modules given by file path."""
        loaded_plugins = []
        for module_name, file_path, klass_or_func in self.include_path:
            try:
                file_path = pathlib.Path(file_path)
                spec = importlib.util.spec_from_file_location(module_name, file_path)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                loaded_plugins.append((module_name, getattr(module, klass_or_func)))
            except Exception as exception:
                logging.warning(
                    f"Failed to load plugin module={module_name}, class={klass_or_func} at {file_path}",
                    exc_info=exception,
                )
                continue
        return loaded_plugins

    def _CollectBuiltinGenerators(
        self, builtin_generators: list[tuple[str, str, str]]
    ) -> list[tuple[str, type[ACLGenerator]]]:
        """Import built-in modules by name."""
        loaded_generators = []
        for target, module_name, class_or_func in builtin_generators:
            if self.disable_builtin and (
                module_name in self.disable_builtin or target in self.disable_builtin
            ):
                continue
            try:
                module = import_module(module_name)
                loaded_generators.append((target, getattr(module, class_or_func)))
            except Exception as exception:
                logging.warning(
                    f"Failed to load built-in generator module={module_name}, class={class_or_func}",
                    exc_info=exception,
                )
                continue
        return loaded_generators



# Copyright 2015 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Arista generator."""

from aerleon.lib import cisco


class Error(cisco.Error):
    """Base error class."""


class UnsupportedEosAccessListError(Error):
    """When a filter type is not supported in an EOS policy target."""


class Arista(cisco.Cisco):
    """An Arista policy object.

    EOS devices differ slightly from Cisco, omitting the extended argument to
    ACLs for example. There are other items such as 'tracked' we may want to add
    in the future.
    """

    _PLATFORM = 'arista'
    SUFFIX = '.eacl'
    # Protocols should be emitted as they were in the policy (names).
    _PROTO_INT = False

    # Arista omits the "extended" access-list argument.
    def _AppendTargetByFilterType(self, filter_name: str, filter_type: str) -> list[str]:
        """Takes in the filter name and type and appends headers.

        Args:
          filter_name: Name of the current filter
          filter_type: Type of current filter

        Returns:
          list of strings

        Raises:
          UnsupportedEosAccessListError: When unknown filter type is used.
        """
        target = []
        if filter_type == 'standard':
            if filter_name.isdigit():
                target.append(f'no access-list {filter_name}')
            else:
                target.append(f'no ip access-list standard {filter_name}')
                target.append(f'ip access-list standard {filter_name}')
        elif filter_type == 'extended':
            target.append(f'no ip access-list {filter_name}')
            target.append(f'ip access-list {filter_name}')
        elif filter_type == 'object-group':
            target.append(f'no ip access-list {filter_name}')
            target.append(f'ip access-list {filter_name}')
        elif filter_type == 'inet6':
            target.append(f'no ipv6 access-list {filter_name}')
            target.append(f'ipv6 access-list {filter_name}')
        else:
            raise UnsupportedEosAccessListError(
                f'access list type {filter_type} not supported by {self._PLATFORM}'
            )
        return target



# Copyright 2014 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Juniper SRX generator for loopback ACLs.

This is a subclass of Juniper generator. Juniper SRX loopback filter
uses the same syntax as regular Juniper stateless ACLs, with minor
differences. This subclass effects those differences.
"""

from aerleon.lib import juniper


class Term(juniper.Term):
    """Single SRXlo term representation."""

    _PLATFORM = 'srxlo'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.term.protocol = ['icmp6' if x == 'icmpv6' else x for x in self.term.protocol]
        self.term.protocol_except = [
            'icmp6' if x == 'icmpv6' else x for x in self.term.protocol_except
        ]

    def NormalizeIcmpTypes(
        self, icmp_types: list[str], protocols: list[str], af: str
    ) -> list[int]:
        protocols = ['icmpv6' if x == 'icmp6' else x for x in protocols]
        return super().NormalizeIcmpTypes(icmp_types, protocols, af)


class SRXlo(juniper.Juniper):
    """SRXlo generator."""

    _PLATFORM = 'srxlo'
    SUFFIX = '.jsl'
    _TERM = Term

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()
        # flexible match is MX/Trio only
        supported_tokens.remove('flexible_match_range')
        # currently only support 'encapsulate' in juniper
        supported_tokens.remove('encapsulate')
        # currently only support 'port-mirror' in juniper
        supported_tokens.remove('port_mirror')
        return supported_tokens, supported_sub_tokens



# Copyright 2020 Google LLC
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Juniper MS-MPC  generator for Aerleon."""

from absl import logging

from aerleon.lib import aclgenerator, juniper, nacaddr, policy

MAX_IDENTIFIER_LEN = 55  # It is really 63, but leaving room for added chars


class Term(juniper.Term):
    """Representation of an individual Juniper MS-MPC term.

    The __str__ method must be implemented.

    Args: term policy.Term object

    """

    _PLATFORM = 'msmpc'
    _DEFAULT_INDENT = 20
    _NO_APPLY_GROUPS_INDENT = 12
    _ACTIONS = {'accept': 'accept', 'deny': 'discard', 'reject': 'reject'}
    # msmpc supports a limited number of protocol names
    # https://www.juniper.net/documentation/us/en/software/junos/security-policies/topics/ref/statement/applications-edit-protocol.html
    _SUPPORTED_PROTOCOL_NAMES = (
        'ah',
        'egp',
        'esp',
        'gre',
        'icmp',
        'icmpv6',
        'igmp',
        'ipip',
        #'node', A pseudo-protocol which may require additional handling
        'ospf',
        'pim',
        'rsvp',
        'sctp',
        'tcp',
        'udp',
    )

    def __init__(
        self,
        term: policy.Term,
        term_type: str,
        noverbose: bool,
        filter_name: str,
        apply_groups: bool,
    ) -> None:
        enable_dsmo = False
        super().__init__(term, term_type, enable_dsmo, noverbose)
        self.term = term
        self.term_type = term_type
        self.noverbose = noverbose
        self.apply_groups = apply_groups
        self.filter_name = filter_name

        for prot in self.term.protocol:
            if prot not in self._SUPPORTED_PROTOCOL_NAMES:
                loc = self.term.protocol.index(prot)
                self.term.protocol[loc] = str(self.PROTO_MAP.get(prot, prot))

    def __str__(self) -> str:
        if self.enable_dsmo:
            raise NotImplementedError('enable_dsmo not implemented for msmpc')

        if self.apply_groups:
            indent = self._DEFAULT_INDENT
        else:
            indent = self._NO_APPLY_GROUPS_INDENT
        ret_str = juniper.Config(indent=indent)

        # COMMENTS
        # this deals just fine with multi line comments, but we could probably
        # output them a little cleaner; do things like make sure the
        # len(output) < 80, etc. Note, if 'noverbose' is set for the filter, skip
        # all comment processing.
        if not self.noverbose:
            if self.term.owner:
                self.term.comment.append(f'Owner: {self.term.owner}')
            if self.term.comment:
                ret_str.Append('/*')
                for comment in self.term.comment:
                    for line in comment.split('\n'):
                        ret_str.Append(f"** {line}")
                ret_str.Append('*/')

        # Term verbatim output - this will skip over normal term creation
        # code.  Warning generated from policy.py if appropriate.
        if self.term.verbatim:
            for next_term in self.term.verbatim:
                if next_term[0] == self._PLATFORM:
                    ret_str.Append(str(next_term[1]), verbatim=True)
            return str(ret_str)

        # Determine whether there are any match conditions for the term.
        has_match_criteria = (
            self.term.address
            or self.term.dscp_except
            or self.term.dscp_match
            or self.term.destination_address
            or self.term.destination_port
            or self.term.destination_prefix
            or self.term.destination_prefix_except
            or self.term.encapsulate
            or self.term.ether_type
            or self.term.flexible_match_range
            or self.term.forwarding_class
            or self.term.forwarding_class_except
            or self.term.fragment_offset
            or self.term.hop_limit
            or self.term.next_ip
            or self.term.port
            or self.term.precedence
            or self.term.protocol
            or self.term.protocol_except
            or self.term.source_address
            or self.term.source_port
            or self.term.source_prefix
            or self.term.source_prefix_except
            or self.term.traffic_type
            or self.term.ttl
        )

        suffixes = []
        duplicate_term = False
        has_icmp = 'icmp' in self.term.protocol
        has_icmpv6 = 'icmpv6' in self.term.protocol
        has_v4_ip = (
            self.term.GetAddressOfVersion('source_address', self.AF_MAP.get('inet'))
            or self.term.GetAddressOfVersion('source_address_exclude', self.AF_MAP.get('inet'))
            or self.term.GetAddressOfVersion('destination_address', self.AF_MAP.get('inet'))
            or self.term.GetAddressOfVersion(
                'destination_address_exclude', self.AF_MAP.get('inet')
            )
        )
        has_v6_ip = (
            self.term.GetAddressOfVersion('source_address', self.AF_MAP.get('inet6'))
            or self.term.GetAddressOfVersion('source_address_exclude', self.AF_MAP.get('inet6'))
            or self.term.GetAddressOfVersion('destination_address', self.AF_MAP.get('inet6'))
            or self.term.GetAddressOfVersion(
                'destination_address_exclude', self.AF_MAP.get('inet6')
            )
        )

        if self.term_type == 'mixed':
            if not (has_v4_ip or has_v6_ip):
                suffixes = ['inet']
            elif not has_v6_ip:
                suffixes = ['inet']
            elif not has_v4_ip:
                suffixes = ['inet6']
            else:
                suffixes = ['inet', 'inet6']
                duplicate_term = True
        if not suffixes and self.term_type in ['inet', 'inet6']:
            suffixes = [self.term_type]

        for suffix in suffixes:
            if (
                self.term_type == 'mixed'
                and (not (has_icmp and has_icmpv6))
                and (has_v4_ip and has_v6_ip)
            ):
                if (has_icmp and suffix != 'inet') or (has_icmpv6 and suffix != 'inet6'):
                    continue
            source_address = self.term.GetAddressOfVersion(
                'source_address', self.AF_MAP.get(suffix)
            )
            source_address_exclude = self.term.GetAddressOfVersion(
                'source_address_exclude', self.AF_MAP.get(suffix)
            )
            source_address, source_address_exclude = self._MinimizePrefixes(
                source_address, source_address_exclude
            )
            destination_address = self.term.GetAddressOfVersion(
                'destination_address', self.AF_MAP.get(suffix)
            )
            destination_address_exclude = self.term.GetAddressOfVersion(
                'destination_address_exclude', self.AF_MAP.get(suffix)
            )
            destination_address, destination_address_exclude = self._MinimizePrefixes(
                destination_address, destination_address_exclude
            )
            if (
                (not source_address)
                and self.term.GetAddressOfVersion('source_address', self.AF_MAP.get('mixed'))
                and not source_address_exclude
            ) or (
                (not destination_address)
                and self.term.GetAddressOfVersion('destination_address', self.AF_MAP.get('mixed'))
                and not destination_address_exclude
            ):
                continue
            if (
                (has_icmpv6 and not has_icmp and suffix == 'inet')
                or (has_icmp and not has_icmpv6 and suffix == 'inet6')
            ) and self.term_type != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_PROTO.substitute(
                        term=self.term.name, proto=', '.join(self.term.protocol), af=suffix
                    )
                )
                return ''

            # NAME
            # if the term is inactive we have to set the prefix
            if self.term.inactive:
                term_prefix = 'inactive:'
            else:
                term_prefix = ''

            ret_str.Append(
                '%s term %s%s {'
                % (term_prefix, self.term.name, f"-{suffix}" if duplicate_term else '')
            )

            # We only need a "from {" clause if there are any conditions to match.
            if has_match_criteria:
                ret_str.Append('from {')
                # SOURCE ADDRESS
                if source_address or source_address_exclude:
                    ret_str.Append('source-address {')
                    if source_address:
                        for saddr in source_address:
                            for comment in self._Comment(saddr):
                                ret_str.Append(f'{comment}')
                            if saddr.version == 6 and 0 < saddr.prefixlen < 16:
                                for saddr2 in saddr.subnets(new_prefix=16):
                                    ret_str.Append(f'{saddr2};')
                            else:
                                if saddr == nacaddr.IPv6('0::0/0'):
                                    saddr = 'any-ipv6'
                                elif saddr == nacaddr.IPv4('0.0.0.0/0'):
                                    saddr = 'any-ipv4'
                                ret_str.Append(f'{saddr};')

                    # SOURCE ADDRESS EXCLUDE
                    if source_address_exclude:
                        for ex in source_address_exclude:
                            for comment in self._Comment(ex):
                                ret_str.Append(f'{comment}')
                            if ex.version == 6 and 0 < ex.prefixlen < 16:
                                for ex2 in ex.subnets(new_prefix=16):
                                    ret_str.Append(f'{ex2} except;')
                            else:
                                if ex == nacaddr.IPv6('0::0/0'):
                                    ex = 'any-ipv6'
                                elif ex == nacaddr.IPv4('0.0.0.0/0'):
                                    ex = 'any-ipv4'
                                ret_str.Append(f'{ex} except;')
                    ret_str.Append('}')  # source-address {...}

                # DESTINATION ADDRESS
                if destination_address or destination_address_exclude:
                    ret_str.Append('destination-address {')
                    if destination_address:
                        for daddr in destination_address:
                            for comment in self._Comment(daddr):
                                ret_str.Append(f'{comment}')
                            if daddr.version == 6 and 0 < daddr.prefixlen < 16:
                                for daddr2 in daddr.subnets(new_prefix=16):
                                    ret_str.Append(f'{daddr2};')
                            else:
                                if daddr == nacaddr.IPv6('0::0/0'):
                                    daddr = 'any-ipv6'
                                elif daddr == nacaddr.IPv4('0.0.0.0/0'):
                                    daddr = 'any-ipv4'
                                ret_str.Append(f'{daddr};')

                    # DESTINATION ADDRESS EXCLUDE
                    if destination_address_exclude:
                        for ex in destination_address_exclude:
                            for comment in self._Comment(ex):
                                ret_str.Append(f'{comment}')
                            if ex.version == 6 and 0 < ex.prefixlen < 16:
                                for ex2 in ex.subnets(new_prefix=16):
                                    ret_str.Append(f'{ex2} except;')
                            else:
                                if ex == nacaddr.IPv6('0::0/0'):
                                    ex = 'any-ipv6'
                                elif ex == nacaddr.IPv4('0.0.0.0/0'):
                                    ex = 'any-ipv4'
                                ret_str.Append(f'{ex} except;')
                    ret_str.Append('}')  # destination-address {...}

                # source prefix <except> list
                if self.term.source_prefix or self.term.source_prefix_except:
                    for pfx in self.term.source_prefix:
                        ret_str.Append(f"source-prefix-list {pfx};")
                    for epfx in self.term.source_prefix_except:
                        ret_str.Append(f"source-prefix-list {epfx} except;")

                # destination prefix <except> list
                if self.term.destination_prefix or self.term.destination_prefix_except:
                    for pfx in self.term.destination_prefix:
                        ret_str.Append(f"destination-prefix-list {pfx};")
                    for epfx in self.term.destination_prefix_except:
                        ret_str.Append(f"destination-prefix-list {epfx} except;")

                # APPLICATION
                if (
                    self.term.source_port
                    or self.term.destination_port
                    or self.term.icmp_type
                    or self.term.protocol
                ):
                    if hasattr(self.term, 'replacement_application_name'):
                        ret_str.Append(
                            f"application-sets {self.term.replacement_application_name}-app;"
                        )
                    else:
                        ret_str.Append(
                            'application-sets '
                            + self.filter_name[: ((MAX_IDENTIFIER_LEN) // 2)]
                            + self.term.name[-((MAX_IDENTIFIER_LEN) // 2) :]
                            + '-app;'
                        )
                ret_str.Append('}')  # from {...}

            ret_str.Append('then {')
            # ACTION
            for action in self.term.action:
                ret_str.Append(f"{self._ACTIONS.get(str(action))};")
            if self.term.logging and 'disable' not in [x.value for x in self.term.logging]:
                ret_str.Append('syslog;')
            ret_str.Append('}')  # then {...}
            ret_str.Append('}')  # term {...}
        return str(ret_str)


class JuniperMSMPC(aclgenerator.ACLGenerator):
    """Juniper MSMPC rendering class.

    This class takes a policy object and renders output into
    a syntax which is understood ny Juniper routers with MS-MPC cards.

    Args:
      pol: policy.Policy object
    """

    _PLATFORM = 'msmpc'
    SUFFIX = '.msmpc'
    _SUPPORTED_AF = frozenset(('inet', 'inet6', 'mixed'))
    _AF_MAP = {'inet': 4, 'inet6': 6, 'mixed': None}
    _AF_ICMP_MAP = {'icmp': 'inet', 'icmpv6': 'inet6'}
    _SUPPORTED_DIRECTION = {
        '': 'input-output',
        'ingress': 'input',
        'egress': 'output',
    }

    _OPTIONAL_SUPPORTED_KEYWORDS = frozenset(
        [
            'expiration',
        ]
    )

    def __init__(self, pol: policy.Policy, exp_info: int):
        self.applications = {}
        super().__init__(pol, exp_info)

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {
            'destination_prefix',
            'destination_prefix_except',
            'icmp_code',
            'logging',
            'owner',
            'source_prefix',
            'source_prefix_except',
        }
        supported_sub_tokens.update(
            {
                'option': {
                    'established',
                    # TODO(sneakywombat): add all options to lex.
                    '.*',  # make ArbitraryOptions work, yolo.
                    'tcp-established',
                    'inactive',
                }
            }
        )
        return supported_tokens, supported_sub_tokens

    def _BuildPort(self, ports: list[tuple[int, int]]):
        """Transform specified ports into list and ranges.

        Args:
          ports: a policy terms list of ports

        Returns:
          port_list: list of ports and port ranges
        """
        port_list = []
        for p in ports:
            if p[0] == p[1]:
                port_list.append(str(p[0]))
            else:
                port_list.append(f'{str(p[0])}-{str(p[1])}')
        return port_list

    def _GenerateApplications(self, filter_name: str):
        target = []
        apps_set_list = []
        target.append('applications {')
        done_apps = []
        for app in sorted(self.applications[filter_name], key=lambda x: x['name']):
            app_list = []
            if app in done_apps:
                continue

            if app['protocol'] or app['sport'] or app['dport'] or app['icmp-type']:
                # generate ICMP statements
                if app['icmp-type']:
                    if app['timeout']:
                        timeout = app['timeout']
                    else:
                        timeout = 60
                    num_terms = len(app['protocol']) * len(app['icmp-type'])
                    apps_set_list.append(f"application-set {app['name']}-app {{")
                    for i in range(num_terms):
                        apps_set_list.append(f"application {app['name']}{'-app%d' % (i + 1)};")
                    apps_set_list.append('}')  # application-set {...}

                    term_counter = 0
                    for i, code in enumerate(app['icmp-type']):
                        for proto in app['protocol']:
                            target.append(
                                f"application {app['name']}{'-app%d' % (term_counter + 1)} {{"
                            )
                            if proto == 'icmp':
                                target.append(f'application-protocol {proto};')
                            target.append(f'protocol {proto};')
                            target.append(f'{proto}-type {str(code)};')
                            if app['icmp-code']:
                                target.append(f"{proto}-code {self._Group(app['icmp-code'])};")
                            if int(timeout):
                                target.append(f'inactivity-timeout {int(timeout)};')
                            target.append('}')  # application {...}
                            term_counter += 1
                # generate non-ICMP statements
                else:
                    i = 1
                    apps_set_list.append(f"application-set {app['name']}-app {{")

                    for proto in app['protocol'] or ['']:
                        for sport in app['sport'] or ['']:
                            for dport in app['dport'] or ['']:
                                chunks = []
                                if proto:
                                    chunks.append(f'protocol {proto};')
                                if sport and ('udp' in proto or 'tcp' in proto):
                                    chunks.append(f'source-port {sport};')
                                if dport and ('udp' in proto or 'tcp' in proto):
                                    chunks.append(f'destination-port {dport};')
                                if app['timeout']:
                                    chunks.append(f" inactivity-timeout {int(app['timeout'])};")
                                if chunks:
                                    apps_set_list.append(
                                        f"application {app['name']}{'-app%d;' % i}"
                                    )
                                    app_list.append(f"application {app['name']}{'-app%d {' % i}")
                                    for chunk in chunks:
                                        app_list.append(chunk)
                                    app_list.append('}')
                                    i += 1
                    apps_set_list.append('}')

                done_apps.append(app)
                if app_list:
                    for item in app_list:
                        target.append(item)

        for item in apps_set_list:
            target.append(item)
        target.append('}')
        # Return the output only if there is content inside of
        # the "applications {\n}" lines, otherwise return nothing.
        if len(target) > 2:
            return target
        else:
            return []

    def _TranslatePolicy(self, pol: policy.Policy, exp_info: int):
        self.junipermsmpc_policies = []
        for header, terms in pol.filters:
            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)
            filter_options.remove(filter_name)
            apply_groups = 'no-apply-groups' not in filter_options
            filter_direction = None
            filter_type = None
            noverbose = 'noverbose' in filter_options
            self.applications[filter_name] = []

            if not apply_groups:
                filter_options.remove('no-apply-groups')

            if noverbose:
                # noverbose is a strict boolean, remove it
                # prior to iterating through the other options
                # that require additional processing.
                filter_options.remove('noverbose')

            for filter_opt in filter_options:
                # validate address families
                if filter_opt in self._SUPPORTED_AF:
                    if not filter_type:
                        filter_type = filter_opt
                        continue
                    else:
                        raise ConflictingTargetOptionsError(
                            'only one address family can be '
                            'specified per header "%s"' % ' '.join(filter_options)
                        )
                # validate direction
                if filter_opt in self._SUPPORTED_DIRECTION.keys():
                    if not filter_direction:
                        filter_direction = self._SUPPORTED_DIRECTION.get(filter_opt)
                        continue
                    else:
                        raise ConflictingTargetOptionsError(
                            'only one direction can be '
                            'specified per header "%s"' % ' '.join(filter_options)
                        )
                raise UnsupportedHeaderError(
                    'MSMPC Generator currently does not support '
                    '%s as a header option "%s"' % (filter_opt, ' '.join(filter_options))
                )

            if not filter_direction:
                filter_direction = self._SUPPORTED_DIRECTION.get('')
            if not filter_type:
                filter_type = 'mixed'

            term_names = set()
            new_terms = []
            for term in terms:
                # Application sets need to be unique system-wide, so we construct
                # a name from a combination of the filter and term names, shortening
                # to the roughly half of the max identifier length for each part.
                # When shortening, we take the start of the filter name and the end of
                # the term name in a hope that we omit the most common bits
                # like -inbound and accept-.
                modified_term_name = (
                    filter_name[: ((MAX_IDENTIFIER_LEN) // 2)]
                    + term.name[-((MAX_IDENTIFIER_LEN) // 2) :]
                )
                if term.stateless_reply:
                    logging.warning(
                        'WARNING: Term %s is a stateless reply term and will not be ' 'rendered.',
                        term.name,
                    )
                    continue
                if {'established', 'tcp-established'}.intersection(term.option):
                    logging.warning(
                        'Skipping established term %s because MSMPC is stateful.', term.name
                    )
                    continue
                # if inactive is set, deactivate the term and remove the option.
                if 'inactive' in term.option:
                    term.inactive = True
                    term.option.remove('inactive')
                if term.name in term_names:
                    raise JuniperMSMPCFilterError('Duplicate term name')
                term_names.add(term.name)
                new_term = Term(term, filter_type, noverbose, filter_name, apply_groups)
                new_terms.append(new_term)

                # Because MSMPC terms can contain inet and inet6 addresses. We have to
                # have ability to recover proper AF for ICMP type we need.
                # If protocol is empty or we cannot map to inet or inet6 we insert bogus
                # af_type name which will cause new_term.NormalizeIcmpTypes to fail.
                if not term.protocol:
                    icmp_af_type = 'unknown_af_icmp'
                else:
                    icmp_af_type = self._AF_ICMP_MAP.get(term.protocol[0], 'unknown_af_icmp')
                tmp_icmptype = new_term.NormalizeIcmpTypes(
                    term.icmp_type, term.protocol, icmp_af_type
                )
                # NormalizeIcmpTypes returns [''] for empty, convert to [] for eval
                normalized_icmptype = tmp_icmptype if tmp_icmptype != [''] else []
                # rewrites the protocol icmpv6 to icmp6
                if 'icmpv6' in term.protocol:
                    protocol = list(term.protocol)
                    protocol[protocol.index('icmpv6')] = 'icmp6'
                else:
                    protocol = term.protocol
                # MSMPC requires tcp and udp to specify ports, rather than imply all
                # ports
                if 'udp' in term.protocol or 'tcp' in term.protocol:
                    if not term.source_port and not term.destination_port:
                        term.destination_port = [[1, 65535]]
                new_application_set = {
                    'sport': self._BuildPort(term.source_port),
                    'dport': self._BuildPort(term.destination_port),
                    'protocol': protocol,
                    'icmp-type': normalized_icmptype,
                    'icmp-code': term.icmp_code,
                    'timeout': term.timeout,
                }

                for application_set in self.applications[filter_name]:
                    if all(
                        item in list(application_set.items())
                        for item in new_application_set.items()
                    ):
                        new_application_set = ''
                        term.replacement_application_name = application_set['name']
                        break
                    if (
                        modified_term_name == application_set['name']
                        and new_application_set != application_set
                    ):
                        raise ConflictingApplicationSetsError(
                            f'Application set {modified_term_name} has a conflicting entry'
                        )

                if new_application_set:
                    new_application_set['name'] = modified_term_name
                    self.applications[filter_name].append(new_application_set)

            self.junipermsmpc_policies.append(
                (header, filter_name, filter_direction, new_terms, apply_groups)
            )

    def _Group(self, group: list[str], lc: bool = True):
        """If 1 item return it, else return [ item1 item2 ].

        Args:
          group: a list.  could be a list of strings (protocols) or a list of tuples
            (ports)
          lc: return a lower cased result for text.  Default is True.

        Returns:
          rval: a string surrounded by '[' and '];' if len(group) > 1
                or with just ';' appended if len(group) == 1
        """

        def _FormattedGroup(el: list[str], lc: bool = True):
            """Return the actual formatting of an individual element.

            Args:
              el: either a string (protocol) or a tuple (ports)
              lc: return lower cased result for text.  Default is True.

            Returns:
              string: either the lower()'ed string or the ports, hyphenated
                      if they're a range, or by itself if it's not.
            """
            if isinstance(el, str):
                if not lc:
                    return el
                else:
                    return el.lower()
            elif isinstance(el, int):
                return str(el)
            # type is a tuple below here
            elif el[0] == el[1]:
                return '%d' % el[0]
            else:
                return '%d-%d' % (el[0], el[1])

        if len(group) > 1:
            rval = f"[ {' '.join([_FormattedGroup(x, lc=lc) for x in group])} ];"
        else:
            rval = f"{_FormattedGroup(group[0], lc=lc)};"
        return rval

    def __str__(self) -> str:
        target = juniper.Config()
        for (
            header,
            filter_name,
            filter_direction,
            terms,
            apply_groups,
        ) in self.junipermsmpc_policies:
            if apply_groups:
                target.Append('groups {')
                target.Append('replace:')
            target.Append('/*')

            # we want the acl to contain id and date tags, but p4 will expand
            # the tags here when we submit the generator, so we have to trick
            # p4 into not knowing these words.  like taking c-a-n-d-y from a
            # baby.
            for line in aclgenerator.AddRepositoryTags('** '):
                target.Append(line)
            target.Append('**')

            for comment in header.comment:
                for line in comment.split('\n'):
                    target.Append(f"** {line}")
            target.Append('*/')

            if apply_groups:
                target.Append('%s {' % filter_name)
            target.Append('services {')
            target.Append('stateful-firewall {')
            target.Append('rule %s {' % filter_name)
            target.Append(f'match-direction {filter_direction};')
            for term in terms:
                term_str = str(term)
                if term_str:
                    target.Append(term_str, verbatim=True)
            target.Append('}')  # rule { ... }
            target.Append('}')  # stateful-firewall { ... }
            target.Append('}')  # services { ... }
            for line in self._GenerateApplications(filter_name):
                target.Append(line)
            if apply_groups:
                target.Append('}')  # filter_name { ... }
                target.Append('}')  # groups { ... }
                target.Append(f'apply-groups {filter_name};')
        return f"{target!s}\n"


class Error(juniper.Error):
    pass


class JuniperMSMPCFilterError(Error):
    pass


class ConflictingApplicationSetsError(Error):
    pass


class ConflictingTargetOptionsError(Error):
    pass


class UnsupportedHeaderError(Error):
    pass



import ipaddress
import itertools as it
import random


def write_excludes_testcase(ipstr, excludelist='', max_prefix_range=8, max_random_subnets=30):
    """
    Writes a testcase to the tests/utils/address_exclude_test_cases.txt file.
    Note that the number of prefixes to consider grows exponentially, so unless
    you *do* want to consider a large pool to randomly select from, keep it at the default

    Args:
      ipstr: the ip network as a string (v4 or v6) to base the test on.
      excludelist: optional comma-separated string of ip networks to exclude
      max_prefix_range: the largest number of prefixes to consider.
      max_random_subnets: the number of subnets to do exclusion tests for, if randomly generating
    Returns:
      None
    """
    ip = ipaddress.ip_network(ipstr)
    if len(excludelist) == 0:  # empty excludelist, making a random one
        prefixrange = min(max_prefix_range, ip.max_prefixlen - ip.prefixlen)
        excludelist = it.chain.from_iterable(ip.subnets(i) for i in range(1, prefixrange + 1))
        total_ips = 2**prefixrange
        ip_positions = set(random.choices(range(total_ips), k=min(max_random_subnets, total_ips)))
        compress_map = (1 if i in ip_positions else 0 for i in range(total_ips))
        excludelist = list(it.compress(excludelist, compress_map))
    else:
        excludelist = list(map(ipaddress.ip_network, excludelist.split(',')))

    result_list = []
    for address in excludelist:
        result_list.append(ip.address_exclude(address))

    ipst = str(ip)
    exst = ",".join(map(str, excludelist))
    rest = ";".join(",".join(map(str, sorted(result))) for result in result_list)
    with open('tests/utils/address_exclude_test_cases.txt', 'a') as f:
        f.write(f"{ipst} {exst} {rest}\n")



# Copyright 2011 Capirca Project Authors All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Cisco ASA renderer."""

import ipaddress
import re
from typing import Union, cast

from absl import logging

from aerleon.lib import aclgenerator, cisco, nacaddr, summarizer
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Policy, Term

_ACTION_TABLE = {
    'accept': 'permit',
    'deny': 'deny',
    'reject': 'deny',
    'next': '! next',
    'reject-with-tcp-rst': 'deny',  # tcp rst not supported
}


# generic error class
class Error(aclgenerator.Error):
    """Generic error class."""


class UnsupportedCiscoAccessListError(Error):
    """Raised when we're give a non named access list."""


class StandardAclTermError(Error):
    """Raised when there is a problem in a standard access list."""


class NoCiscoPolicyError(Error):
    """Raised when a policy is errantly passed to this module for rendering."""


class Term(cisco.ExtendedTerm):
    """A single ACL Term."""

    def __init__(
        self, term: Term, filter_name: str, af: int = 4, enable_dsmo: bool = False
    ) -> None:
        self.term = term
        self.filter_name = filter_name
        self.options = []
        assert af in (4, 6)
        self.af = af
        self.enable_dsmo = enable_dsmo

    def __str__(self) -> str:
        ret_str = ['\n']

        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (self.af == 6 and 'icmp' in self.term.protocol) or (
            self.af == 4 and 'icmpv6' in self.term.protocol
        ):
            ret_str.append(f'remark Term {self.term.name}')
            ret_str.append('remark not rendered due to protocol/AF mismatch.')
            return '\n'.join(ret_str)

        ret_str.append(f'access-list {self.filter_name} remark {self.term.name}')
        if self.term.owner:
            self.term.comment.append(f'Owner: {self.term.owner}')
        for comment in self.term.comment:
            for line in comment.split('\n'):
                ret_str.append(f'access-list {self.filter_name} remark {str(line)[:100]}')

        # Term verbatim output - this will skip over normal term creation
        # code by returning early.  Warnings provided in policy.py.
        if self.term.verbatim:
            for line in self.term.verbatim:
                if line[0] == 'ciscoasa':
                    ret_str.append(str(line[1]))
            return '\n'.join(ret_str)

        # protocol
        if not self.term.protocol:
            protocol = ['ip']
        else:
            # fix the protocol
            protocol = self.term.protocol

        # source address
        if self.term.source_address:
            source_address = self.term.GetAddressOfVersion('source_address', self.af)
            source_address_exclude = self.term.GetAddressOfVersion(
                'source_address_exclude', self.af
            )
            if source_address_exclude:
                source_address = nacaddr.ExcludeAddrs(source_address, source_address_exclude)
            if self.enable_dsmo and len(source_address) > 1:
                source_address = summarizer.Summarize(source_address)
        else:
            # source address not set
            source_address = ['any']

        # destination address
        if self.term.destination_address:
            destination_address = self.term.GetAddressOfVersion('destination_address', self.af)
            destination_address_exclude = self.term.GetAddressOfVersion(
                'destination_address_exclude', self.af
            )
            if destination_address_exclude:
                destination_address = nacaddr.ExcludeAddrs(
                    destination_address, destination_address_exclude
                )
            if self.enable_dsmo and len(destination_address) > 1:
                destination_address = summarizer.Summarize(destination_address)
        else:
            # destination address not set
            destination_address = ['any']

        # options
        extra_options = []
        for opt in [str(x) for x in self.term.option]:
            if opt.find('tcp-established') == 0 and 6 in protocol:
                extra_options.append('established')
            elif opt.find('established') == 0 and 6 in protocol:
                # only needed for TCP, for other protocols policy.py handles high-ports
                extra_options.append('established')
        self.options.extend(extra_options)

        # ports
        source_port = [()]
        destination_port = [()]
        if self.term.source_port:
            source_port = self.term.source_port
        if self.term.destination_port:
            destination_port = self.term.destination_port

        # logging
        if self.term.logging:
            self.options.append('log')
            if 'disable' in [x.value for x in self.term.logging]:
                self.options.append('disable')

        # icmp-types
        icmp_types = ['']
        if self.term.icmp_type:
            icmp_types = self.NormalizeIcmpTypes(self.term.icmp_type, self.term.protocol, self.af)

        for saddr in source_address:
            for daddr in destination_address:
                for sport in source_port:
                    for dport in destination_port:
                        for proto in protocol:
                            for icmp_type in icmp_types:
                                # only output address family appropriate IP addresses
                                do_output = False
                                if self.af == 4:
                                    if (
                                        (isinstance(saddr, nacaddr.IPv4)) or (saddr == 'any')
                                    ) and ((isinstance(daddr, nacaddr.IPv4)) or (daddr == 'any')):
                                        do_output = True
                                    elif isinstance(saddr, summarizer.DSMNet) or isinstance(
                                        daddr, summarizer.DSMNet
                                    ):
                                        do_output = True
                                if self.af == 6:
                                    if (
                                        (isinstance(saddr, nacaddr.IPv6)) or (saddr == 'any')
                                    ) and ((isinstance(daddr, nacaddr.IPv6)) or (daddr == 'any')):
                                        do_output = True
                                if do_output:
                                    ret_str.extend(
                                        self._TermletToStr(
                                            self.filter_name,
                                            _ACTION_TABLE.get(str(self.term.action[0])),
                                            proto,
                                            saddr,
                                            sport,
                                            daddr,
                                            dport,
                                            icmp_type,
                                            self.options,
                                        )
                                    )

        return '\n'.join(ret_str)

    def _TermletToStr(
        self,
        filter_name: str,
        action: str,
        proto: str,
        saddr: Union[str, IPv4, IPv6],
        sport: Union[tuple[()], tuple[int, int]],
        daddr: Union[str, IPv4, IPv6],
        dport: Union[tuple[()], tuple[int, int]],
        icmp_type: str,
        option: list[str],
    ) -> list[str]:
        """Take the various compenents and turn them into a cisco acl line.

        Args:
          filter_name: name of the filter
          action: str, action
          proto: str, protocl
          saddr: str or ipaddress, source address
          sport: str list or none, the source port
          daddr: str or ipaddress, the destination address
          dport: str list or none, the destination port
          icmp_type: icmp-type numeric specification (if any)
          option: list of options, eg. 'logging' tokens.

        Returns:
          string of the cisco acl line, suitable for printing.
        """
        # inet4
        if isinstance(saddr, nacaddr.IPv4) or isinstance(saddr, ipaddress.IPv4Network):
            saddr = cast(self.IPV4_ADDRESS, saddr)
            if saddr.num_addresses > 1:
                saddr = f'{saddr.network_address} {saddr.netmask}'
            else:
                saddr = f'host {saddr.network_address}'
        if isinstance(daddr, nacaddr.IPv4) or isinstance(daddr, ipaddress.IPv4Network):
            daddr = cast(self.IPV4_ADDRESS, daddr)
            if daddr.num_addresses > 1:
                daddr = f'{daddr.network_address} {daddr.netmask}'
            else:
                daddr = f'host {daddr.network_address}'
        if isinstance(saddr, summarizer.DSMNet):
            saddr = '%s %s' % summarizer.ToDottedQuad(saddr, negate=False)

        if isinstance(daddr, summarizer.DSMNet):
            daddr = '%s %s' % summarizer.ToDottedQuad(daddr, negate=False)
        # inet6
        if isinstance(saddr, nacaddr.IPv6) or isinstance(saddr, ipaddress.IPv6Network):
            saddr = cast(self.IPV6_ADDRESS, saddr)
            if saddr.num_addresses > 1:
                saddr = f'{saddr.network_address}/{saddr.prefixlen}'
            else:
                saddr = f'host {saddr.network_address}'
        if isinstance(daddr, nacaddr.IPv6) or isinstance(daddr, ipaddress.IPv6Network):
            daddr = cast(self.IPV6_ADDRESS, daddr)
            if daddr.num_addresses > 1:
                daddr = f'{daddr.network_address}/{daddr.prefixlen}'
            else:
                daddr = f'host {daddr.network_address}'

        # fix ports
        if not sport:
            sport = ''
        elif sport[0] != sport[1]:
            sport = ' range {} {}'.format(
                cisco.PortMap.GetProtocol(sport[0], proto),
                cisco.PortMap.GetProtocol(sport[1], proto),
            )
        else:
            sport = f' eq {cisco.PortMap.GetProtocol(sport[0], proto)}'

        if not dport:
            dport = ''
        elif dport[0] != dport[1]:
            dport = ' range {} {}'.format(
                cisco.PortMap.GetProtocol(dport[0], proto),
                cisco.PortMap.GetProtocol(dport[1], proto),
            )
        else:
            dport = f' eq {cisco.PortMap.GetProtocol(dport[0], proto)}'

        if not option:
            option = ['']

        # Prevent UDP from appending 'established' to ACL line
        sane_options = list(option)
        if proto == 'udp' and 'established' in sane_options:
            sane_options.remove('established')

        ret_lines = []

        # str(icmp_type) is needed to ensure 0 maps to '0' instead of FALSE
        icmp_type = str(cisco.PortMap.GetProtocol(icmp_type, 'icmp'))

        ret_lines.append(
            'access-list %s extended  %s %s %s %s %s %s %s %s'
            % (
                filter_name,
                action,
                proto,
                saddr,
                sport,
                daddr,
                dport,
                icmp_type,
                ' '.join(sane_options),
            )
        )

        # remove any trailing spaces and replace multiple spaces with singles
        stripped_ret_lines = [re.sub(r'\s+', ' ', x).rstrip() for x in ret_lines]
        return stripped_ret_lines


class CiscoASA(aclgenerator.ACLGenerator):
    """A cisco ASA policy object."""

    _PLATFORM = 'ciscoasa'
    _DEFAULT_PROTOCOL = 'ip'
    SUFFIX = '.asa'

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {'logging', 'owner'}

        supported_sub_tokens.update(
            {
                'option': {'established', 'tcp-established'},
                # Warning, some of these are mapped
                # differently. See _ACTION_TABLE
                'action': {'accept', 'deny', 'reject', 'next', 'reject-with-tcp-rst'},
            }
        )
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        self.ciscoasa_policies = []
        for header, terms in self.policy.filters:
            filter_name = header.FilterName(self._PLATFORM)
            filter_options = header.FilterOptions(self._PLATFORM)
            enable_dsmo = len(filter_options) > 1 and 'enable_dsmo' in filter_options[1:]

            new_terms = []
            # now add the terms
            for term in terms:
                new_terms.append(str(Term(term, filter_name, enable_dsmo=enable_dsmo)))

            self.ciscoasa_policies.append((header, filter_name, new_terms))

    def __str__(self) -> str:
        target = []

        for header, filter_name, terms in self.ciscoasa_policies:
            target.append(f'clear configure access-list {filter_name}')

            # add the p4 tags
            target.extend(aclgenerator.AddRepositoryTags(f'access-list {filter_name} remark '))

            # add a header comment if one exists
            for comment in header.comment:
                for line in comment.split('\n'):
                    target.append(f'access-list {filter_name} remark {line}')

            # now add the terms
            for term in terms:
                target.append(str(term))

            # end for header, filter_name, filter_type...

        return '\n'.join(target)



# Copyright 2011 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Cisco generator."""

import ipaddress
from typing import Any, Optional, Union, cast

from absl import logging

from aerleon.lib import aclgenerator, addressbook, nacaddr, summarizer
from aerleon.lib.nacaddr import IPv4, IPv6
from aerleon.lib.policy import Policy, Term
from aerleon.lib.summarizer import DSMNet

_ACTION_TABLE = {
    'accept': 'permit',
    'deny': 'deny',
    'reject': 'deny',
    'next': '! next',
    'reject-with-tcp-rst': 'deny',  # tcp rst not supported
}

_COMMENT_MAX_WIDTH = 70


# generic error class
class Error(aclgenerator.Error):
    """Generic error class."""


class CiscoDuplicateTermError(Error):
    """Raised on duplicate term names."""


class CiscoNextIpError(Error):
    """Raised when next-ip is misconfigured."""


class UnsupportedCiscoAccessListError(Error):
    """Raised when we're give a non named access list."""


class StandardAclTermError(Error):
    """Raised when there is a problem in a standard access list."""


class ExtendedACLTermError(Error):
    """Raised when there is a problem in an extended access list."""


class TermStandard:
    """A single standard ACL Term."""

    def __init__(
        self, term: Term, filter_name: str, platform: str = 'cisco', verbose: bool = True
    ) -> None:
        self.term = term
        self.filter_name = filter_name
        self.platform = platform
        self.options = []
        self.logstring = ''
        self.dscpstring = ''
        self.verbose = verbose
        # sanity checking for standard acls
        if self.term.protocol:
            raise StandardAclTermError('Standard ACLs cannot specify protocols')
        if self.term.icmp_type:
            raise StandardAclTermError(
                'ICMP Type specifications are not permissible in standard ACLs'
            )
        if (
            self.term.source_address
            or self.term.source_address_exclude
            or self.term.destination_address
            or self.term.destination_address_exclude
        ):
            raise StandardAclTermError('Standard ACLs cannot use source or destination addresses')
        if self.term.option:
            raise StandardAclTermError('Standard ACLs prohibit use of options')
        if self.term.source_port or self.term.destination_port:
            raise StandardAclTermError('Standard ACLs prohibit use of port numbers')
        if self.term.logging:
            logging.warning(
                'WARNING: Standard ACL logging is set in filter %s, term %s and '
                'may not implemented on all IOS versions',
                self.filter_name,
                self.term.name,
            )
            self.logstring = ' log'
        if self.term.dscp_match:
            logging.warning(
                'WARNING: dscp-match is set in filter %s, term %s and may not be '
                'implemented on all IOS version',
                self.filter_name,
                self.term.name,
            )
            self.dscpstring = f" dscp{self.term.dscp_match}"

    def __str__(self) -> str:
        ret_str = []

        # Term verbatim output - this will skip over normal term creation
        # code by returning early.  Warnings provided in policy.py.
        if self.term.verbatim:
            for next_verbatim in self.term.verbatim:
                if next_verbatim[0] == self.platform:
                    ret_str.append(str(next_verbatim[1]))
            return '\n'.join(ret_str)

        v4_addresses = [x for x in self.term.address if not isinstance(x, nacaddr.IPv6)]
        prefix = ''
        if self.filter_name.isdigit():
            prefix = f'access-list {self.filter_name}'

        if self.verbose:
            ret_str.append(f'{prefix} remark {self.term.name}')
            comments = aclgenerator.WrapWords(self.term.comment, _COMMENT_MAX_WIDTH)
            for comment in comments:
                ret_str.append(f'{prefix} remark {comment}')

        action = _ACTION_TABLE.get(str(self.term.action[0]))
        if v4_addresses:
            for addr in v4_addresses:
                if self.platform == 'arista':
                    if addr.prefixlen == 32:
                        ret_str.append(
                            f'{prefix} {action} host {addr.network_address}{self.logstring}{self.dscpstring}'
                        )
                    else:
                        ret_str.append(
                            f'{prefix} {action} {addr.network_address}/{addr.prefixlen}{self.logstring}{self.dscpstring}'
                        )
                elif addr.prefixlen == 32:
                    ret_str.append(
                        f'{prefix} {action} {addr.network_address}{self.logstring}{self.dscpstring}'
                    )
                else:
                    ret_str.append(
                        f'{prefix} {action} {addr.network_address} {addr.hostmask}{self.logstring}{self.dscpstring}'
                    )
        else:
            ret_str.append(f'{prefix} {action} any{self.logstring}{self.dscpstring}')

        return '\n'.join(ret_str)


class ObjectGroup:
    """Used for printing out the object group definitions.

    since the ports don't store the token name information, we have
    to fudge their names.  ports will be written out like

      object-group port <low_port>-<high_port>
        range <low-port> <high-port>
      exit

    where as the addressess can be written as

      object-group address ipv4 first-term-source-address
        172.16.0.0
        172.20.0.0 255.255.0.0
        172.22.0.0 255.128.0.0
        172.24.0.0
        172.28.0.0
      exit
    """

    def __init__(self) -> None:
        self.filter_name = ''
        self.addressbook = addressbook.Addressbook()
        self.terms = []

    def AddTerm(self, term: Term) -> None:
        if term.source_address:
            self.addressbook.AddAddresses('', term.source_address)
        if term.destination_address:
            self.addressbook.AddAddresses('', term.destination_address)
        self.terms.append(term)

    def AddName(self, filter_name: str) -> None:
        self.filter_name = filter_name

    def __str__(self) -> str:
        ret_str = ['\n']
        # netgroups will contain two-tuples of group name string and family int.
        netgroups = set()
        ports = {}

        # I don't have an easy way get the token name used in the pol file
        # w/o reading the pol file twice (with some other library) or doing
        # some other ugly hackery. Instead, the entire block of source and dest
        # addresses for a given term is given a unique, computable name which
        # is not related to the NETWORK.net token name.  that's what you get
        # for using cisco, which has decided to implement its own meta language.

        # Create network object-groups
        for zone, name, ips, _ in self.addressbook.Walk():
            for version in (4, 6):
                vips = [i for i in ips if i.version == version]
                if vips:
                    ret_str.append(f'object-group network ipv{version} {name}')
                    for ip in vips:
                        ret_str.append(f' {ip.network_address}/{ip.prefixlen}')
                    ret_str.append('exit\n')

            # Create port object-groups
        for term in self.terms:
            for port in term.source_port + term.destination_port:
                if not port:
                    continue
                port_key = f'{port[0]}-{port[1]}'
                if port_key not in ports:
                    ports[port_key] = True
                    ret_str.append(f'object-group port {port_key}')
                    if port[0] != port[1]:
                        ret_str.append(' range %d %d' % (port[0], port[1]))
                    else:
                        ret_str.append(' eq %d' % port[0])
                    ret_str.append('exit\n')
        return '\n'.join(ret_str)


class PortMap:
    """Map port numbers to service names."""

    # Define port mappings common to all protocols
    _PORTS_TCP = {
        179: 'bgp',
        19: 'chargen',
        514: 'cmd',
        13: 'daytime',
        9: 'discard',
        53: 'domain',
        7: 'echo',
        512: 'exec',
        79: 'finger',
        21: 'ftp',
        20: 'ftp-data',
        70: 'gopher',
        443: 'https',
        113: 'ident',
        194: 'irc',
        543: 'klogin',
        544: 'kshell',
        389: 'ldap',
        636: 'ldaps',
        513: 'login',
        515: 'lpd',
        2049: 'nfs',
        119: 'nntp',
        496: 'pim-auto-rp',
        109: 'pop2',
        110: 'pop3',
        1723: 'pptp',
        554: 'rtsp',
        5060: 'sip',
        25: 'smtp',
        22: 'ssh',
        111: 'sunrpc',
        49: 'tacacs',
        517: 'talk',
        23: 'telnet',
        540: 'uucp',
        43: 'whois',
        80: 'www',
    }

    _PORTS_UDP = {
        512: 'biff',
        68: 'bootpc',
        67: 'bootps',
        9: 'discard',
        195: 'dnsix',
        53: 'domain',
        7: 'echo',
        500: 'isakmp',
        434: 'mobile-ip',
        42: 'nameserver',
        138: 'netbios-dgm',
        137: 'netbios-ns',
        2049: 'nfs',
        123: 'ntp',
        496: 'pim-auto-rp',
        520: 'rip',
        161: 'snmp',
        162: 'snmptrap',
        111: 'sunrpc',
        514: 'syslog',
        49: 'tacacs',
        517: 'talk',
        69: 'tftp',
        37: 'time',
        513: 'who',
        177: 'xdmcp',
    }

    _TYPES_ICMP = {
        6: 'alternate-address',
        31: 'conversion-error',
        8: 'echo',
        0: 'echo-reply',
        16: 'information-reply',
        15: 'information-request',
        18: 'mask-reply',
        17: 'mask-request',
        32: 'mobile-redirect',
        12: 'parameter-problem',
        5: 'redirect',
        9: 'router-advertisement',
        10: 'router-solicitation',
        4: 'source-quench',
        11: 'time-exceeded',
        14: 'timestamp-reply',
        13: 'timestamp-request',
        30: 'traceroute',
        3: 'unreachable',
    }

    # Combine cisco-specific port mappings with common ones
    _CISCO_PORTS_TCP = {
        5190: 'aol',
        1494: 'citrix-ica',
        2748: 'ctiqbe',
        1720: 'h323',
        101: 'hostname',
        143: 'imap4',
        750: 'kerberos',
        1352: 'lotusnotes',
        139: 'netbios-ssn',
        5631: 'pcanywhere-data',
        1521: 'sqlnet',
    }
    _CISCO_PORTS_TCP.update(_PORTS_TCP)

    _CISCO_PORTS_UDP = {
        750: 'kerberos',
        5632: 'pcanywhere-status',
        1645: 'radius',
        1646: 'radius-acct',
        5510: 'secureid-udp',
        5060: 'sip',
    }
    _CISCO_PORTS_UDP.update(_PORTS_UDP)

    # Combine arista-specific port mappings with common ones
    _ARISTA_PORTS_TCP = {
        143: 'imap',
        88: 'kerberos',
    }
    _ARISTA_PORTS_TCP.update(_PORTS_TCP)

    _ARISTA_PORTS_UDP = {
        88: 'kerberos',
        1812: 'radius',
        1813: 'radius-acct',
    }
    _ARISTA_PORTS_UDP.update(_PORTS_UDP)

    # Full port map data structure
    _PORT_MAP = {
        'cisco': {'tcp': _CISCO_PORTS_TCP, 'udp': _CISCO_PORTS_UDP, 'icmp': _TYPES_ICMP},
        'arista': {'tcp': _ARISTA_PORTS_TCP, 'udp': _ARISTA_PORTS_UDP, 'icmp': _TYPES_ICMP},
    }

    @staticmethod
    def GetProtocol(port_num: int, proto: str, platform: str = 'cisco'):
        """Converts a port number to a name or returns the number.

        Args:
          port_num: integer representing the port number.
          proto: string representing proto (tcp, udp, etc).
          platform: string representing platform (cisco, arista)

        Returns:
          A name of the protocol or the port number that was provided.
        """
        try:
            port_map = PortMap._PORT_MAP[platform][proto]
            return port_map[port_num]
        except KeyError:
            return port_num


class Term(aclgenerator.Term):
    """This is an general Term object that all other Cisco Terms should inherite from."""

    ALLOWED_PROTO_STRINGS = [
        'eigrp',
        'gre',
        'icmp',
        'igmp',
        'igrp',
        'ip',
        'ipinip',
        'nos',
        'pim',
        'tcp',
        'udp',
        'sctp',
        'ahp',
    ]
    IPV4_ADDRESS = Union[nacaddr.IPv4, ipaddress.IPv4Network]
    IPV6_ADDRESS = Union[nacaddr.IPv6, ipaddress.IPv6Network]

    def __init__(
        self,
        term: Term,
        af: int = 4,
        proto_int: bool = True,
        enable_dsmo: bool = False,
        term_remark: bool = True,
        platform: str = 'cisco',
        verbose: bool = True,
        filter_type: Optional[str] = None,
    ) -> None:
        self.term = term
        self.proto_int = proto_int
        self.options = []
        self.enable_dsmo = enable_dsmo
        self.term_remark = term_remark
        self.platform = platform
        self.verbose = verbose
        self.filter_type = filter_type
        # Our caller should have already verified the address family.
        assert af in (4, 6)
        self.af = af
        if af == 4:
            self.text_af = 'inet'
        else:
            self.text_af = 'inet6'

    def __str__(self) -> str:
        ret_str = ['\n']

        # Don't render icmpv6 protocol terms under inet, or icmp under inet6
        if (
            (self.af == 6 and 'icmp' in self.term.protocol)
            or (self.af == 6 and self.PROTO_MAP['icmp'] in self.term.protocol)
            or (self.af == 4 and 'icmpv6' in self.term.protocol)
            or (self.af == 4 and self.PROTO_MAP['icmpv6'] in self.term.protocol)
        ):
            if self.filter_type != 'mixed':
                logging.warning(
                    self.NO_AF_LOG_PROTO.substitute(
                        term=self.term.name, proto=', '.join(self.term.protocol), af=self.text_af
                    )
                )
            return ''

        # verbose
        if self.verbose:
            comments = []
            if self.term_remark:
                comments.append(self.term.name)
            if self.term.owner:
                comments.append(f'Owner: {self.term.owner}')
            comments.extend(self.term.comment)

            comments = aclgenerator.WrapWords(comments, _COMMENT_MAX_WIDTH)
            if comments and comments[0]:
                for comment in comments:
                    ret_str.append(f' remark {comment!s}')

        # Term verbatim output - this will skip over normal term creation
        # code by returning early.  Warnings provided in policy.py.
        if self.term.verbatim:
            for next_verbatim in self.term.verbatim:
                if next_verbatim[0] == self.platform:
                    ret_str.append(str(next_verbatim[1]))
            return '\n'.join(ret_str)

        # protocol
        if not self.term.protocol:
            if self.af == 6:
                protocol = ['ipv6']
            elif self.platform == 'ciscoxr':
                protocol = ['ipv4']
            else:
                protocol = ['ip']
        elif self.term.protocol == ['hopopt'] or self.term.protocol == self.PROTO_MAP['hopopt']:
            protocol = ['hbh']
        elif self.proto_int:
            protocol = [
                (
                    proto
                    if proto in self.ALLOWED_PROTO_STRINGS or proto.isnumeric()
                    else self.PROTO_MAP.get(proto)
                )
                for proto in self.term.protocol
            ]
        else:
            protocol = self.term.protocol
        # Arista can not process acls with esp/ah, these must appear as integers.
        if self.platform == 'arista':
            if 'esp' in protocol:
                protocol = [x if x != 'esp' else '50' for x in protocol]
            if 'ah' in protocol:
                protocol = [x if x != 'ah' else '51' for x in protocol]

        # addresses
        # source address
        if self.term.source_address:
            source_address = self.term.GetAddressOfVersion('source_address', self.af)
            source_address_exclude = self.term.GetAddressOfVersion(
                'source_address_exclude', self.af
            )
            if source_address_exclude:
                source_address = nacaddr.ExcludeAddrs(source_address, source_address_exclude)
            if not source_address:
                if self.filter_type != 'mixed':
                    logging.warning(
                        self.NO_AF_LOG_ADDR.substitute(
                            term=self.term.name, direction='source', af=self.text_af
                        )
                    )
                return ''
            if self.enable_dsmo:
                source_address = summarizer.Summarize(source_address)
        else:
            # source address not set
            source_address = [nacaddr.IPv4('0.0.0.0/0', token='any')]
        fixed_src_addresses = {x for x in source_address}

        # destination address
        if self.term.destination_address:
            destination_address = self.term.GetAddressOfVersion('destination_address', self.af)
            destination_address_exclude = self.term.GetAddressOfVersion(
                'destination_address_exclude', self.af
            )
            if destination_address_exclude:
                destination_address = nacaddr.ExcludeAddrs(
                    destination_address, destination_address_exclude
                )
            if not destination_address:
                if self.filter_type != 'mixed':
                    logging.warning(
                        self.NO_AF_LOG_ADDR.substitute(
                            term=self.term.name, direction='destination', af=self.text_af
                        )
                    )
                return ''
            if self.enable_dsmo:
                destination_address = summarizer.Summarize(destination_address)
        else:
            # destination address not set
            destination_address = [nacaddr.IPv4('0.0.0.0/0', token='any')]

        fixed_dst_addresses = {x for x in destination_address}

        # ports
        source_port = [()]
        destination_port = [()]

        # source port
        if self.term.source_port:
            source_port = self._FixConsecutivePorts(self.term.source_port)

        # destination port
        if self.term.destination_port:
            destination_port = self._FixConsecutivePorts(self.term.destination_port)

        # options
        opts = [str(x) for x in self.term.option]
        if (self.PROTO_MAP['tcp'] in protocol or 'tcp' in protocol) and (
            'tcp-established' in opts or 'established' in opts
        ):
            if 'established' not in self.options:
                self.options.append('established')
        # Using both 'fragments' and 'is-fragment', ref Github Issue #187
        if ('fragments' in opts) or ('is-fragment' in opts):
            if 'fragments' not in self.options:
                self.options.append('fragments')
        # ACL-based Forwarding
        if (
            (self.platform == 'ciscoxr')
            and not self.term.action
            and self.term.next_ip
            and ('nexthop1' not in opts)
        ):
            if len(self.term.next_ip) > 1:
                raise CiscoNextIpError(
                    f'The following term has more than one next IP value: {self.term.name}'
                )
            if not isinstance(self.term.next_ip[0], nacaddr.IPv4) and not isinstance(
                self.term.next_ip[0], nacaddr.IPv6
            ):
                raise CiscoNextIpError(
                    f'Next IP value must be an IP address. Invalid term: {self.term.name}'
                )
            if self.term.next_ip[0].num_addresses > 1:
                raise CiscoNextIpError(
                    f'The following term has a subnet instead of a host: {self.term.name}'
                )
            nexthop = self.term.next_ip[0].network_address
            nexthop_protocol = 'ipv4' if nexthop.version == 4 else 'ipv6'
            self.options.append(f'nexthop1 {nexthop_protocol} {nexthop}')
            action = _ACTION_TABLE.get('accept')

        # action
        if self.term.action:
            action = _ACTION_TABLE.get(str(self.term.action[0]))

        # logging
        if self.term.logging:
            self.options.append('log')

        # dscp; unlike srx, cisco only supports single, non-except values
        if self.term.dscp_match:
            if len(self.term.dscp_match) > 1:
                raise ExtendedACLTermError(
                    'Extended ACLs cannot specify more than one dscp match value'
                )
            else:
                self.options.append(f"dscp {' '.join(self.term.dscp_match)}")

        # icmp-types
        icmp_types = ['']
        if self.term.icmp_type:
            icmp_types = self.NormalizeIcmpTypes(self.term.icmp_type, self.term.protocol, self.af)
        icmp_codes = ['']
        if self.term.icmp_code:
            icmp_codes = self.term.icmp_code

        fixed_opts = {}
        for p in protocol:
            fixed_opts[p] = self._FixOptions(p, self.options)

        # temlet constructor
        for saddr in sorted(fixed_src_addresses):
            for daddr in sorted(fixed_dst_addresses):
                for sport in sorted(source_port):
                    for dport in sorted(destination_port):
                        for proto in sorted(protocol):
                            opts = fixed_opts[proto]
                            # cisconx uses icmp for both ipv4 and ipv6
                            if self.platform == 'cisconx':
                                if self.af == 6:
                                    proto = 'icmp' if proto == 'icmpv6' else proto
                            for icmp_type in sorted(icmp_types):
                                for icmp_code in sorted(icmp_codes):
                                    ret_str.extend(
                                        self._TermletToStr(
                                            action,
                                            proto,
                                            self._GetIpString(saddr),
                                            self._FormatPort(sport, proto),
                                            self._GetIpString(daddr),
                                            self._FormatPort(dport, proto),
                                            icmp_type,
                                            icmp_code,
                                            opts,
                                        )
                                    )

        return '\n'.join(ret_str)

    def _GetIpString(self, addr: Union[IPv6, IPv4, DSMNet]) -> str:
        """Formats the address object for printing in the ACL.

        Args:
          addr: str or ipaddr, address
        Returns:
          An address string suitable for the ACL.
        """
        # DSMO enabled
        if isinstance(addr, summarizer.DSMNet):
            return '%s %s' % summarizer.ToDottedQuad(addr, negate=True)
        if addr.prefixlen == 0:
            return 'any'
        if isinstance(addr, nacaddr.IPv4) or isinstance(addr, ipaddress.IPv4Network):
            addr = cast(self.IPV4_ADDRESS, addr)
            if addr.num_addresses > 1:
                if self.platform in ('arista', 'cisconx'):
                    return addr.with_prefixlen
                return f'{addr.network_address} {addr.hostmask}'
            if addr.num_addresses == 1 and self.platform == 'cisconx':
                return f'{addr.with_prefixlen}'
            return f'host {addr.network_address}'
        if isinstance(addr, nacaddr.IPv6) or isinstance(addr, ipaddress.IPv6Network):
            addr = cast(self.IPV6_ADDRESS, addr)
            if addr.num_addresses > 1:
                return addr.with_prefixlen
            return f'host {addr.network_address}'
        return addr

    def _FormatPort(self, port: Union[tuple[()], tuple[int, int]], proto: Union[int, str]) -> str:
        """Returns a formatted port string for the range.

        Args:
          port: The port range represented as a tuple..
          proto: String or int representing a protocol.

        Returns:
          A string used to filter a single or range of ports..
        """
        if not port:
            return ''
        port0 = port[0]
        port1 = port[1]
        if self.platform == 'arista':
            port0 = PortMap.GetProtocol(port0, proto, self.platform)
            port1 = PortMap.GetProtocol(port1, proto, self.platform)

        if port[0] != port[1]:
            return f'range {port0} {port1}'
        return f'eq {port0}'

    def _FixOptions(
        self, proto: Union[int, str], option: list[Union[str, Any]]
    ) -> list[Union[str, Any]]:
        """Returns a set of options suitable for the given protocol.

        Fix done:
        - Filter out 'established' for UDP.
        - Filter out 'fragments' for TCP/UDP

        Args:
          proto: str or int, protocol
          option: list or none, optional, eg. 'logging' tokens.
        Returns:
          A list of options suitable for that protocol.
        """
        # Prevent UDP from appending 'established' to ACL line
        sane_options = list(option)
        if (proto == self.PROTO_MAP['udp'] or proto == 'udp') and 'established' in sane_options:
            sane_options.remove('established')
        return sane_options

    def _TermletToStr(
        self,
        action: str,
        proto: Union[int, str],
        saddr: str,
        sport: str,
        daddr: str,
        dport: str,
        icmp_type: Union[int, str],
        icmp_code: Union[int, str],
        option: list[str],
    ) -> list[str]:
        """Take the various compenents and turn them into a cisco acl line.

        Args:
          action: str, action
          proto: str or int, protocol
          saddr: str, source address
          sport: str, the source port
          daddr: str, the destination address
          dport: str, the destination port
          icmp_type: icmp-type numeric specification (if any)
          icmp_code: icmp-code numeric specification (if any)
          option: list of options, eg. 'logging' tokens.

        Returns:
          string of the cisco acl line, suitable for printing.

        Raises:
          UnsupportedCiscoAccessListError: When unknown icmp-types specified
        """
        # str(icmp_type) is needed to ensure 0 maps to '0' instead of FALSE
        icmp_type = str(icmp_type)
        icmp_code = str(icmp_code)
        all_elements = [
            action,
            str(proto),
            saddr,
            sport,
            daddr,
            dport,
            icmp_type,
            icmp_code,
            ' '.join(option),
        ]
        non_empty_elements = [x for x in all_elements if x]
        return [f" {' '.join(non_empty_elements)}"]

    def _FixConsecutivePorts(self, port_list: list[tuple[int, int]]) -> list[tuple[int, int]]:
        """Takes a list of tuples and expands the tuple if the range is two.

            http://www.cisco.com/warp/public/cc/pd/si/casi/ca6000/tech/65acl_wp.pdf

        Args:
          port_list: A list of tuples representing ports.

        Returns:
          list of tuples
        """
        temporary_port_list = []
        for low_port, high_port in port_list:
            if low_port == high_port - 1:
                temporary_port_list.append((low_port, low_port))
                temporary_port_list.append((high_port, high_port))
            else:
                temporary_port_list.append((low_port, high_port))
        return temporary_port_list


class ExtendedTerm(Term):
    """A single ExtendedACL Term."""


class ObjectGroupTerm(Term):
    """An individual term of an object-group'd acl.

    Object Group acls are very similar to extended acls in their
    syntax except they use a meta language with address/service
    definitions.

    eg:

      permit tcp first-term-source-address 179-179 ANY

    where first-term-source-address, ANY and 179-179 are defined elsewhere
    in the acl.
    """

    def _FormatPort(self, port: Union[tuple[()], tuple[int, int]], proto: str) -> str:
        """Returns a formatted port string for the range.

        Args:
          port: str list or none, the port range.
          proto: str representing proto (tcp, udp, etc).

        Returns:
          A string suitable for the ACL.
        """
        if not port:
            return ''
        return f'port-group {port[0]}-{port[1]}'

    def _GetIpString(self, addr: Union[IPv4, IPv6]) -> str:
        """Formats the address object for printing in the ACL.

        Args:
          addr: str or ipaddr, address
        Returns:
          An address string suitable for the ACL.
        """
        if addr.prefixlen == 0:
            return f'{addr.parent_token}'
        return f'net-group {addr.parent_token}'


class Cisco(aclgenerator.ACLGenerator):
    """A cisco policy object."""

    _PLATFORM = 'cisco'
    _DEFAULT_PROTOCOL = 'ip'
    SUFFIX = '.acl'
    # Protocols should be emitted as numbers.
    _PROTO_INT = True
    _TERM_REMARK = True

    def _BuildTokens(self) -> tuple[set[str], dict[str, set[str]]]:
        """Build supported tokens for platform.

        Returns:
          tuple containing both supported tokens and sub tokens
        """
        supported_tokens, supported_sub_tokens = super()._BuildTokens()

        supported_tokens |= {
            'address',
            'restrict_address_family',
            'dscp_match',
            'icmp_code',
            'logging',
            'owner',
        }

        supported_sub_tokens.update(
            {
                'option': {'established', 'tcp-established', 'is-fragment', 'fragments'},
                # Warning, some of these are mapped
                # differently. See _ACTION_TABLE
                'action': {'accept', 'deny', 'reject', 'next', 'reject-with-tcp-rst'},
            }
        )
        return supported_tokens, supported_sub_tokens

    def _TranslatePolicy(self, pol: Policy, exp_info: int) -> None:
        self.cisco_policies = []
        # a mixed filter outputs both ipv4 and ipv6 acls in the same output file
        good_filters = ['extended', 'standard', 'object-group', 'inet6', 'mixed', 'enable_dsmo']

        for header, terms in pol.filters:
            obj_target = ObjectGroup()

            filter_options = header.FilterOptions(self._PLATFORM)
            filter_name = header.FilterName(self._PLATFORM)

            self.verbose = True
            if 'noverbose' in filter_options:
                filter_options.remove('noverbose')
                self.verbose = False

            # extended is the most common filter type.
            filter_type = 'extended'
            if len(filter_options) > 1:
                filter_type = filter_options[1]

            # check if filter type is renderable
            if filter_type not in good_filters:
                raise UnsupportedCiscoAccessListError(
                    'access list type %s not supported by %s (good types: %s)'
                    % (filter_type, self._PLATFORM, str(good_filters))
                )

            filter_list = [filter_type]
            if filter_type == 'mixed':
                # Loop through filter and generate output for inet and inet6 in sequence
                filter_list = ['extended', 'inet6']

            for next_filter in filter_list:
                # Numeric access lists can be extended or standard, but have specific
                # known ranges.
                if next_filter == 'extended' and filter_name.isdigit():
                    if int(filter_name) in list(range(1, 100)) + list(range(1300, 2000)):
                        raise UnsupportedCiscoAccessListError(
                            'Access lists between 1-99 and 1300-1999 are reserved for '
                            'standard ACLs'
                        )
                if next_filter == 'standard' and filter_name.isdigit():
                    if int(filter_name) not in list(range(1, 100)) + list(range(1300, 2000)):
                        raise UnsupportedCiscoAccessListError(
                            'Standard access lists must be numeric in the range of 1-99'
                            ' or 1300-1999.'
                        )

                term_dup_check = set()
                new_terms = []
                for term in terms:
                    if term.name in term_dup_check:
                        raise CiscoDuplicateTermError(f'You have a duplicate term: {term.name}')
                    term_dup_check.add(term.name)

                    term.name = self.FixTermLength(term.name)
                    af = 'inet'
                    if next_filter == 'inet6':
                        af = 'inet6'
                    term = self.FixHighPorts(term, af=af)
                    if not term:
                        continue

                    # Ignore if the term is for a different AF
                    if term.restrict_address_family and term.restrict_address_family != af:
                        continue

                    # render terms based on filter type
                    if next_filter == 'standard':
                        # keep track of sequence numbers across terms
                        new_terms.append(
                            TermStandard(term, filter_name, self._PLATFORM, self.verbose)
                        )
                    elif next_filter == 'extended':
                        enable_dsmo = (
                            len(filter_options) > 2 and filter_options[2] == 'enable_dsmo'
                        )
                        new_terms.append(
                            ExtendedTerm(
                                term,
                                proto_int=self._PROTO_INT,
                                enable_dsmo=enable_dsmo,
                                term_remark=self._TERM_REMARK,
                                platform=self._PLATFORM,
                                verbose=self.verbose,
                                filter_type=filter_type,
                            )
                        )
                    elif next_filter == 'object-group':
                        obj_target.AddTerm(term)
                        new_terms.append(self._GetObjectGroupTerm(term, verbose=self.verbose))
                    elif next_filter == 'inet6':
                        new_terms.append(
                            ExtendedTerm(
                                term,
                                6,
                                proto_int=self._PROTO_INT,
                                platform=self._PLATFORM,
                                verbose=self.verbose,
                                filter_type=filter_type,
                            )
                        )

                # cisco requires different name for the v4 and v6 acls
                if filter_type == 'mixed' and next_filter == 'inet6':
                    filter_name = f'ipv6-{filter_name}'
                self.cisco_policies.append(
                    (header, filter_name, [next_filter], new_terms, obj_target)
                )

    def _GetObjectGroupTerm(self, term: Term, verbose: bool = True) -> ObjectGroupTerm:
        """Returns an ObjectGroupTerm object."""
        return ObjectGroupTerm(term, verbose=verbose)

    def _AppendTargetByFilterType(self, filter_name: str, filter_type: str) -> list[str]:
        """Takes in the filter name and type and appends headers.

        Args:
          filter_name: Name of the current filter
          filter_type: Type of current filter

        Returns:
          list of strings

        Raises:
          UnsupportedCiscoAccessListError: When unknown filter type is used.
        """
        target = []
        if filter_type == 'standard':
            if filter_name.isdigit():
                target.append(f'no access-list {filter_name}')
            else:
                target.append(f'no ip access-list standard {filter_name}')
                target.append(f'ip access-list standard {filter_name}')
        elif filter_type == 'extended':
            target.append(f'no ip access-list extended {filter_name}')
            target.append(f'ip access-list extended {filter_name}')
        elif filter_type == 'object-group':
            target.append(f'no ip access-list extended {filter_name}')
            target.append(f'ip access-list extended {filter_name}')
        elif filter_type == 'inet6':
            target.append(f'no ipv6 access-list {filter_name}')
            target.append(f'ipv6 access-list {filter_name}')
        else:
            raise UnsupportedCiscoAccessListError(
                f'access list type {filter_type} not supported by {self._PLATFORM}'
            )
        return target

    def _RepositoryTagsHelper(
        self, target: Optional[list[str]] = None, filter_type: str = '', filter_name: str = ''
    ) -> list[str]:
        if target is None:
            target = []
        if filter_type == 'standard' and filter_name.isdigit():
            target.extend(
                aclgenerator.AddRepositoryTags(
                    f'access-list {filter_name} remark ', date=False, revision=False
                )
            )
        else:
            target.extend(aclgenerator.AddRepositoryTags(' remark ', date=False, revision=False))
        return target

    def __str__(self) -> str:
        target_header = []
        target = []
        # add the p4 tags
        target.extend(aclgenerator.AddRepositoryTags('! '))

        for header, filter_name, filter_list, terms, obj_target in self.cisco_policies:
            for filter_type in filter_list:
                target.extend(self._AppendTargetByFilterType(filter_name, filter_type))
                if filter_type == 'object-group':
                    obj_target.AddName(filter_name)

                # Add the Perforce Id/Date tags, these must come after
                # remove/re-create of the filter, otherwise config mode doesn't
                # know where to place these remarks in the configuration.
                if self.verbose:
                    target = self._RepositoryTagsHelper(target, filter_type, filter_name)

                    # add a header comment if one exists

                    for comment in aclgenerator.WrapWords(header.comment, _COMMENT_MAX_WIDTH):
                        for line in comment.split('\n'):
                            if (
                                self._PLATFORM == 'cisco'
                                and filter_type == 'standard'
                                and filter_name.isdigit()
                            ):
                                target.append(f'access-list {filter_name} remark {line}')
                            else:
                                target.append(f' remark {line}')

                # now add the terms
                for term in terms:
                    term_str = str(term)
                    if term_str:
                        target.append(term_str)

            if obj_target.addressbook.GetZoneNames():
                target = [str(obj_target)] + target
            # ensure that the header is always first
            target = target_header + target
            target += ['', 'exit', '']
        return '\n'.join(target)



# Copyright 2022 Google Inc. All Rights Reserved.
# Modifications Copyright 2022-2023 Aerleon Project Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Juniper EVO generator.

This is a subclass of Juniper generator. Juniper EVO software (Junos EVO)
uses the same syntax as regular Juniper (Junos) ACLs, with minor differences.
This subclass effects those differences.
"""

from aerleon.lib import aclgenerator, juniper


class Term(juniper.Term):
    """Single Juniper EVO term representation."""

    _PLATFORM = 'juniperevo'
    _INGRESS = 'ingress'
    _EGRESS = 'egress'
    _INET6 = 'inet6'
    _PROTOCOL = 'protocol'
    _PROTOCOL_EXCEPT = 'protocol-except'
    _NEXT_HEADER = 'next-header'
    _NEXT_HEADER_EXCEPT = 'next-header-except'
    _PAYLOAD_PROTOCOL = 'payload-protocol'
    _PAYLOAD_PROTOCOL_EXCEPT = 'payload-protocol-except'

    def __str__(self):
        self._Ipv6ProtocolMatch()
        term_config = super().__str__()
        # Reset to original syntax.
        self._TERM_TYPE[self._INET6][self._PROTOCOL] = self._NEXT_HEADER
        self._TERM_TYPE[self._INET6][self._PROTOCOL_EXCEPT] = self._NEXT_HEADER_EXCEPT
        return term_config

    def _Ipv6ProtocolMatch(self):
        """Use the correct syntax to match protocols after the IPv6 header.

        Refer to juniperevo.md in documentation for matching syntax.

        Returns:
          None

        Raises:
          FilterDirectionError: If a direction is not provided for the filter
            e.g. ingress or egress
        """
        self.extension_headers = ['hop-by-hop', 'fragment']
        # 'hopopt' is renamed to 'hop-by-hop' in juniper base class, add an
        # additional key with the same protocol number to aid renaming.
        self.PROTO_MAP['hop-by-hop'] = 0

        if self.term_type == self._INET6:
            if self.filter_direction != self._INGRESS and self.filter_direction != self._EGRESS:
                raise FilterDirectionError(
                    'a direction must be specified for Junos '
                    'EVO IPv6 filter; this is required to '
                    'render the correct syntax when matching '
                    'protocols headers that follow the IPv6 '
                    'header'
                )

            # Default to rendering filter for physical interfaces.
            if self.interface_type is None:
                self.interface_type = 'physical'

            # Ingress filter.
            if self.filter_direction == self._INGRESS:
                if self.interface_type == 'physical':
                    if not any(header in self.term.protocol for header in self.extension_headers):
                        self._TERM_TYPE[self._INET6][self._PROTOCOL] = self._PAYLOAD_PROTOCOL

                    if not any(
                        header in self.term.protocol_except for header in self.extension_headers
                    ):
                        self._TERM_TYPE[self._INET6][
                            self._PROTOCOL_EXCEPT
                        ] = self._PAYLOAD_PROTOCOL_EXCEPT

                if self.interface_type == 'loopback':
                    self._TERM_TYPE[self._INET6][self._PROTOCOL] = self._PAYLOAD_PROTOCOL
                    self._TERM_TYPE[self._INET6][
                        self._PROTOCOL_EXCEPT
                    ] = self._PAYLOAD_PROTOCOL_EXCEPT

                    self.term.protocol = aclgenerator.ProtocolNameToNumber(
                        self.term.protocol, self.extension_headers, self.PROTO_MAP
                    )

                    self.term.protocol_except = aclgenerator.ProtocolNameToNumber(
                        self.term.protocol_except, self.extension_headers, self.PROTO_MAP
                    )

            # Egress filter.
            if self.filter_direction == self._EGRESS:
                self._TERM_TYPE[self._INET6][self._PROTOCOL] = self._PAYLOAD_PROTOCOL
                self._TERM_TYPE[self._INET6][self._PROTOCOL_EXCEPT] = self._PAYLOAD_PROTOCOL_EXCEPT

                self.term.protocol = aclgenerator.ProtocolNameToNumber(
                    self.term.protocol, self.extension_headers, self.PROTO_MAP
                )

                self.term.protocol_except = aclgenerator.ProtocolNameToNumber(
                    self.term.protocol_except, self.extension_headers, self.PROTO_MAP
                )


class JuniperEvo(juniper.Juniper):
    """Juniper EVO generator."""

    _PLATFORM = 'juniperevo'
    SUFFIX = '.evojcl'
    _TERM = Term


class Error(juniper.Error):
    pass


class FilterDirectionError(Error):
    pass
